@@ -48,6 +48,10 @@
 --
 --   * 'SDouble': IEEE-754 double-precision floating point values
 --
+--   * 'SChar', 'SString', 'RegExp': Characters, strings and regular expressions
+--
+--   * 'SList': Symbolic lists (which can be nested)
+--
 --   * 'SArray', 'SFunArray': Flat arrays of symbolic values.
 --
 --   * Symbolic polynomials over GF(2^n), polynomial arithmetic, and CRCs.
@@ -59,7 +63,7 @@
 --
 -- The user can construct ordinary Haskell programs using these types, which behave
 -- very similar to their concrete counterparts. In particular these types belong to the
--- standard classes 'Num', 'Bits', custom versions of 'Eq' ('EqSymbolic') 
+-- standard classes 'Num', 'Bits', custom versions of 'Eq' ('EqSymbolic')
 -- and 'Ord' ('OrdSymbolic'), along with several other custom classes for simplifying
 -- programming with symbolic values. The framework takes full advantage of Haskell's type
 -- inference to avoid many common mistakes.
@@ -88,7 +92,7 @@
 --
 --   * ABC from University of Berkeley: <http://www.eecs.berkeley.edu/~alanmi/abc/>
 --
---   * CVC4 from New York University and University of Iowa: <http://cvc4.cs.nyu.edu/>
+--   * CVC4 from Stanford: <http://cvc4.cs.stanford.edu/web/>
 --
 --   * Boolector from Johannes Kepler University: <http://fmv.jku.at/boolector/>
 --
@@ -116,105 +120,131 @@
 {-# OPTIONS_GHC -fno-warn-orphans  #-}
 
 module Data.SBV (
-  -- * Programming with symbolic values
   -- $progIntro
 
-  -- ** Symbolic types
+  -- * Symbolic types
 
-  -- *** Symbolic bit
-    SBool
-  -- *** Unsigned symbolic bit-vectors
+  -- ** Booleans
+    SBool, oneIf
+  -- *** The Boolean class
+  , Boolean(..)
+  -- *** Logical operations
+  , bAnd, bOr, bAny, bAll
+  -- ** Bit-vectors
+  -- *** Unsigned bit-vectors
   , SWord8, SWord16, SWord32, SWord64
-  -- *** Signed symbolic bit-vectors
+  -- *** Signed bit-vectors
   , SInt8, SInt16, SInt32, SInt64
-  -- *** Signed unbounded integers
+  -- ** Unbounded integers
   -- $unboundedLimitations
   , SInteger
-  -- *** Floating point numbers
+  -- ** Floating point numbers
   -- $floatingPoints
   , SFloat, SDouble
-  -- *** Signed algebraic reals
+  -- ** Algebraic reals
   -- $algReals
   , SReal, AlgReal, sRealToSInteger
+  -- ** Characters, Strings and Regular Expressions
+  -- $strings
+  , SChar, SString
+  -- ** Symbolic lists
+  -- $lists
+  , SList
+  -- * Arrays of symbolic values
+  , SymArray(newArray_, newArray, readArray, writeArray), SArray, SFunArray
 
-  -- ** Creating a symbolic variable
+  -- * Creating symbolic values
+  -- ** Single value
   -- $createSym
-  , sBool, sWord8, sWord16, sWord32, sWord64, sInt8, sInt16, sInt32, sInt64, sInteger, sReal, sFloat, sDouble
+  , sBool, sWord8, sWord16, sWord32, sWord64, sInt8, sInt16, sInt32, sInt64, sInteger, sReal, sFloat, sDouble, sChar, sString, sList
 
-  -- ** Creating a list of symbolic variables
+  -- ** List of values
   -- $createSyms
-  , sBools, sWord8s, sWord16s, sWord32s, sWord64s, sInt8s, sInt16s, sInt32s, sInt64s, sIntegers, sReals, sFloats, sDoubles
-
-  -- *** Abstract SBV type
-  , SBV, HasKind(..), Kind(..)
-  -- *** Arrays of symbolic values
-  , SymArray(..), SArray, SFunArray, mkSFunArray
-
-  -- ** Operations on symbolic values
-  -- *** Word level
-  , sShiftLeft, sShiftRight, sRotateLeft, sRotateRight, sSignedShiftArithRight, sFromIntegral, oneIf
-  , label
-
-  -- *** Exponentiation
-  , (.^)
-  -- *** Splitting, joining, and extending
-  , Splittable(..)
-
-  -- ** Conditionals: Mergeable values
-  , Mergeable(..), ite, iteLazy
-
-  -- ** Symbolic integral numbers
-  , SIntegral
-  -- ** Symbolic finite bits
-  , SFiniteBits(..)
-  -- ** Division
-  , SDivisible(..)
-  -- ** The Boolean class
-  , Boolean(..)
-  -- *** Generalizations of boolean operations
-  , bAnd, bOr, bAny, bAll
-
-  -- * Uninterpreted sorts, constants, and functions
-  -- $uninterpreted
-  , Uninterpreted(..), addAxiom
+  , sBools, sWord8s, sWord16s, sWord32s, sWord64s, sInt8s, sInt16s, sInt32s, sInt64s, sIntegers, sReals, sFloats, sDoubles, sChars, sStrings, sLists
 
   -- * Symbolic Equality and Comparisons
   , EqSymbolic(..), OrdSymbolic(..), Equality(..)
+  -- * Conditionals: Mergeable values
+  , Mergeable(..), ite, iteLazy
 
-  -- * Constraints
-  -- $constrainIntro
-  , constrain, namedConstraint
-  -- ** Cardinality constraints
-  -- $cardIntro
-  , pbAtMost, pbAtLeast, pbExactly, pbLe, pbGe, pbEq, pbMutexed, pbStronglyMutexed
+  -- * Symbolic integral numbers
+  , SIntegral
+  -- * Division and Modulus
+  , SDivisible(..)
+  -- * Bit-vector operations
+  -- ** Conversions
+  , sFromIntegral
+  -- ** Shifts and rotates
+  -- $shiftRotate
+  , sShiftLeft, sShiftRight, sRotateLeft, sRotateRight, sSignedShiftArithRight
+  -- ** Finite bit-vector operations
+  , SFiniteBits(..)
+  -- ** Splitting, joining, and extending
+  , Splittable(..)
+  -- ** Exponentiation
+  , (.^)
+  -- * IEEE-floating point numbers
+  , IEEEFloating(..), RoundingMode(..), SRoundingMode, nan, infinity, sNaN, sInfinity
+  -- ** Rounding modes
+  , sRoundNearestTiesToEven, sRoundNearestTiesToAway, sRoundTowardPositive, sRoundTowardNegative, sRoundTowardZero, sRNE, sRNA, sRTP, sRTN, sRTZ
+  -- ** Conversion to/from floats
+  , IEEEFloatConvertable(..)
+  -- ** Bit-pattern conversions
+  , sFloatAsSWord32, sWord32AsSFloat, sDoubleAsSWord64, sWord64AsSDouble, blastSFloat, blastSDouble
 
   -- * Enumerations
   -- $enumerations
   , mkSymbolicEnumeration
 
-  -- * Properties, proofs, satisfiability, and safety
+  -- * Uninterpreted sorts, constants, and functions
+  -- $uninterpreted
+  , Uninterpreted(..), addAxiom
+
+  -- * Properties, proofs, and satisfiability
   -- $proveIntro
   -- $noteOnNestedQuantifiers
   -- $multiIntro
-  , Predicate, Goal, Provable(..)
-  -- ** Checking safety
+  , Predicate, Goal, Provable(..), solve
+  -- * Constraints
+  -- $constrainIntro
+  -- ** General constraints
+  -- $generalConstraints
+  , constrain, softConstrain
+
+  -- ** Constraint Vacuity
+  -- $constraintVacuity
+
+  -- ** Named constraints and attributes
+  -- $namedConstraints
+  , namedConstraint, constrainWithAttribute
+
+  -- ** Unsat cores
+  -- $unsatCores
+
+  -- ** Cardinality constraints
+  -- $cardIntro
+  , pbAtMost, pbAtLeast, pbExactly, pbLe, pbGe, pbEq, pbMutexed, pbStronglyMutexed
+
+  -- * Checking safety
   -- $safeIntro
   , sAssert, isSafe, SExecutable(..)
-  -- ** Satisfying a sequence of boolean conditions
-  , solve
 
-  -- ** Quick-checking
+  -- * Quick-checking
   , sbvQuickCheck
 
-  -- * Running a symbolic computation
-  , runSMT, runSMTWith
-
-  -- * Solver exceptions
-  , SMTException(..)
-
   -- * Optimization
   -- $optiIntro
-  , OptimizeStyle(..), Penalty(..), Objective(..), minimize, maximize, assertSoft
+
+  -- ** Multiple optimization goals
+  -- $multiOpt
+  , OptimizeStyle(..)
+  -- ** Objectives
+  , Objective(..), Metric(..)
+  -- ** Soft assumptions
+  -- $softAssertions
+  , assertWithPenalty , Penalty(..)
+  -- ** Field extensions
+  -- | If an optimization results in an infinity/epsilon value, the returned `CW` value will be in the corresponding extension field.
   , ExtCW(..), GeneralizedCW(..)
 
   -- * Model extraction
@@ -222,27 +252,33 @@ module Data.SBV (
 
   -- ** Inspecting proof results
   -- $resultTypes
-  , ThmResult(..), SatResult(..), AllSatResult(..), SafeResult(..), OptimizeResult(..), SMTResult(..)
+  , ThmResult(..), SatResult(..), AllSatResult(..), SafeResult(..), OptimizeResult(..), SMTResult(..), SMTReasonUnknown(..)
 
-  -- * IEEE-floating point numbers
-  , IEEEFloating(..), IEEEFloatConvertable(..), RoundingMode(..), SRoundingMode, nan, infinity, sNaN, sInfinity
-  -- ** Rounding modes
-  , sRoundNearestTiesToEven, sRoundNearestTiesToAway, sRoundTowardPositive, sRoundTowardNegative, sRoundTowardZero, sRNE, sRNA, sRTP, sRTN, sRTZ
-  -- ** Bit-pattern conversions
-  , sFloatAsSWord32, sWord32AsSFloat, sDoubleAsSWord64, sWord64AsSDouble, blastSFloat, blastSDouble
+  -- ** Observing expressions
+  -- $observeInternal
+  , observe
 
   -- ** Programmable model extraction
   -- $programmableExtraction
   , SatModel(..), Modelable(..), displayModels, extractModels
   , getModelDictionaries, getModelValues, getModelUninterpretedValues
 
-  -- * SMT Interface: Configurations and solvers
+  -- * SMT Interface
   , SMTConfig(..), Timing(..), SMTLibVersion(..), Solver(..), SMTSolver(..)
-  , boolector, cvc4, yices, z3, mathSAT, abc, defaultSolverConfig, defaultSMTCfg, sbvCheckSolverInstallation, sbvAvailableSolvers
-  , setLogic, setOption, setInfo, setTimeOut
+  -- ** Controlling verbosity
+  -- $verbosity
 
-  -- * Symbolic computations
-  , Symbolic, output, SymWord(..)
+  -- ** Solvers
+  , boolector, cvc4, yices, z3, mathSAT, abc
+  -- ** Configurations
+  , defaultSolverConfig, defaultSMTCfg, sbvCheckSolverInstallation, sbvAvailableSolvers
+  , setLogic, Logic(..), setOption, setInfo, setTimeOut
+  -- ** SBV exceptions
+  , SBVException(..)
+
+  -- * Abstract SBV type
+  , SBV, HasKind(..), Kind(..), SymWord(..)
+  , Symbolic, label, output, runSMT, runSMTWith
 
   -- * Module exports
   -- $moduleExportIntro
@@ -276,15 +312,16 @@ import Data.Word
 import qualified Language.Haskell.TH as TH
 import Data.Generics
 
-import Data.SBV.SMT.Utils (SMTException(..))
-import Data.SBV.Control.Utils (SMTValue)
+import Data.SBV.SMT.Utils (SBVException(..))
+import Data.SBV.Control.Utils (SMTValue (..))
+import Data.SBV.Control.Types (SMTReasonUnknown(..), Logic(..))
 
 -- | Form the symbolic conjunction of a given list of boolean conditions. Useful in expressing
 -- problems with constraints, like the following:
 --
 -- @
---   do [x, y, z] <- sIntegers [\"x\", \"y\", \"z\"]
---      solve [x .> 5, y + z .< x]
+--   sat $ do [x, y, z] <- sIntegers [\"x\", \"y\", \"z\"]
+--            solve [x .> 5, y + z .< x]
 -- @
 solve :: [SBool] -> Symbolic SBool
 solve = return . bAnd
@@ -404,11 +441,9 @@ executable within Haskell, without the need for any custom interpreters. (They a
 Haskell programs, not AST's built out of pieces of syntax.) This provides for an integrated
 feel of the system, one of the original design goals for SBV.
 
-= Incremental mode: Queries
-
-SBV provides a wide variety of ways to utilize SMT-solvers, without requiring the user to
+Incremental query mode: SBV provides a wide variety of ways to utilize SMT-solvers, without requiring the user to
 deal with the solvers themselves. While this mode is convenient, advanced users might need
-access to the underlying solver, using the SMTLib language. For such use cases, SBV allows
+access to the underlying solver at a lower level. For such use cases, SBV allows
 users to have an interactive session: The user can issue commands to the solver, inspect
 the values/results, and formulate new constraints. This advanced feature is available through
 the "Data.SBV.Control" module, where most SMTLib features are made available via a typed-API.
@@ -424,12 +459,13 @@ to communicate with arbitrary SMT solvers.
 -}
 
 {- $multiIntro
+=== Using multiple solvers
 On a multi-core machine, it might be desirable to try a given property using multiple SMT solvers,
 using parallel threads. Even with machines with single-cores, threading can be helpful if you
 want to try out multiple-solvers but do not know which one would work the best
 for the problem at hand ahead of time.
 
-The functions in this section allow proving/satisfiability-checking with multiple
+SBV allows proving/satisfiability-checking with multiple
 backends at the same time. Each function comes in two variants, one that
 returns the results from all solvers, the other that returns the fastest one.
 
@@ -472,27 +508,27 @@ What happens if we make sure to arrange for this invariant? Consider this versio
 
 >>> let safeSub x y = ite (x .>= y) (sub x y) 0
 
-Clearly, 'safeSub' must be safe. And indeed, SBV can prove that:
+Clearly, @safeSub@ must be safe. And indeed, SBV can prove that:
 
 >>> safe (safeSub :: SInt8 -> SInt8 -> SInt8)
 [sub: x >= y must hold!: No violations detected]
 
-Note how we used 'sub' and 'safeSub' polymorphically. We only need to monomorphise our types when a proof
+Note how we used @sub@ and @safeSub@ polymorphically. We only need to monomorphise our types when a proof
 attempt is done, as we did in the 'safe' calls.
 
-If required, the user can pass a 'CallStack' through the first argument to 'sAssert', which will be used
+If required, the user can pass a @CallStack@ through the first argument to 'sAssert', which will be used
 by SBV to print a diagnostic info to pinpoint the failure.
 
-Also see "Data.SBV.Examples.Misc.NoDiv0" for the classic div-by-zero example.
+Also see "Documentation.SBV.Examples.Misc.NoDiv0" for the classic div-by-zero example.
 -}
 
 
 {- $optiIntro
-  SBV can optimize metric functions, i.e., those that generate both bounded 'SIntN', 'SWordN', and unbounded 'SInteger'
+  SBV can optimize metric functions, i.e., those that generate both bounded @SIntN@, @SWordN@, and unbounded 'SInteger'
   types, along with those produce 'SReal's. That is, it can find models satisfying all the constraints while minimizing
   or maximizing user given metrics. Currently, optimization requires the use of the z3 SMT solver as the backend,
   and a good review of these features is given
-  in this paper: <http://www.easychair.org/publications/download/Z_-_Maximal_Satisfaction_with_Z3>.
+  in this paper: <http://www.microsoft.com/en-us/research/wp-content/uploads/2016/02/nbjorner-scss2014.pdf>.
 
   Goals can be lexicographically (default), independently, or pareto-front optimized. The relevant functions are:
 
@@ -502,7 +538,7 @@ Also see "Data.SBV.Examples.Misc.NoDiv0" for the classic div-by-zero example.
   Goals can be optimized at a regular or an extended value: An extended value is either positive or negative infinity
   (for unbounded integers and reals) or positive or negative epsilon differential from a real value (for reals).
 
-  For instance, a call of the form 
+  For instance, a call of the form
 
        @ 'minimize' "name-of-goal" $ x + 2*y @
 
@@ -539,7 +575,14 @@ Optimal model:
   As usual, the programmatic API can be used to extract the values of objectives and model-values ('getModelObjectives',
   'getModelAssignment', etc.) to access these values and program with them further.
 
-== Multiple optimization goals
+  The following examples illustrate the use of basic optimization routines:
+
+     * "Documentation.SBV.Examples.Optimization.LinearOpt": Simple linear-optimization example.
+     * "Documentation.SBV.Examples.Optimization.Production": Scheduling machines in a shop
+     * "Documentation.SBV.Examples.Optimization.VM": Scheduling virtual-machines in a data-center
+-}
+
+{- $multiOpt
 
   Multiple goals can be specified, using the same syntax. In this case, the user gets to pick what style of
   optimization to perform, by passing the relevant 'OptimizeStyle' as the first argument to 'optimize'.
@@ -553,32 +596,38 @@ Optimal model:
     * ['Pareto']. Finally, the user can query for pareto-fronts. A pareto front is an model such that no goal can be made
       "better" without making some other goal "worse."
 
+      Pareto fronts only make sense when the objectives are bounded. If there are unbounded objective values, then the
+      backend solver can loop infinitely. (This is what z3 does currently.) If you are not sure the objectives are
+      bounded, you should first use 'Independent' mode to ensure the objectives are bounded, and then switch to
+      pareto-mode to extract them further.
+
       The optional number argument to 'Pareto' specifies the maximum number of pareto-fronts the user is asking
-      to get. If 'Nothing', SBV will query for all pareto-fronts. Note that pareto-fronts can be infinite
-      in number, so if 'Nothing' is used, there is a potential for infinitely waiting for the SBV-solver interaction
-      to finish. (If you suspect this might be the case, run in 'verbose' mode to see the interaction and
-      put a limiting factor appropriately.)
+      to get. If 'Nothing', SBV will query for all pareto-fronts. Note that pareto-fronts can be really large,
+      so if 'Nothing' is used, there is a potential for waiting indefinitely for the SBV-solver interaction to finish. (If
+      you suspect this might be the case, run in 'verbose' mode to see the interaction and put a limiting factor
+      appropriately.)
+-}
 
-== Soft Assertions
+{- $softAssertions
 
-  Related to optimization, SBV implements soft-asserts via 'assertSoft' calls. A soft assertion
+  Related to optimization, SBV implements soft-asserts via 'assertWithPenalty' calls. A soft assertion
   is a hint to the SMT solver that we would like a particular condition to hold if **possible*.
   That is, if there is a solution satisfying it, then we would like it to hold, but it can be violated
   if there is no way to satisfy it. Each soft-assertion can be associated with a numeric penalty for
   not satisfying it, hence turning it into an optimization problem.
 
-  Note that 'assertSoft' works well with optimization goals ('minimize'/'maximize' etc.),
+  Note that 'assertWithPenalty' works well with optimization goals ('minimize'/'maximize' etc.),
   and are most useful when we are optimizing a metric and thus some of the constraints
   can be relaxed with a penalty to obtain a good solution. Again
-  see <http://www.easychair.org/publications/download/Z_-_Maximal_Satisfaction_with_Z3>
+  see <http://www.microsoft.com/en-us/research/wp-content/uploads/2016/02/nbjorner-scss2014.pdf>
   for a good overview of the features in Z3 that SBV is providing the bridge for.
 
   A soft assertion can be specified in one of the following three main ways:
 
        @
-         'assertSoft' "bounded_x" (x .< 5) 'DefaultPenalty'
-         'assertSoft' "bounded_x" (x .< 5) ('Penalty' 2.3 Nothing)
-         'assertSoft' "bounded_x" (x .< 5) ('Penalty' 4.7 (Just "group-1")) @
+         'assertWithPenalty' "bounded_x" (x .< 5) 'DefaultPenalty'
+         'assertWithPenalty' "bounded_x" (x .< 5) ('Penalty' 2.3 Nothing)
+         'assertWithPenalty' "bounded_x" (x .< 5) ('Penalty' 4.7 (Just "group-1")) @
 
   In the first form, we are saying that the constraint @x .< 5@ must be satisfied, if possible,
   but if this constraint can not be satisfied to find a model, it can be violated with the default penalty of 1.
@@ -588,13 +637,6 @@ Optimal model:
   Finally in the third case, we are also associating this constraint with a group. The group
   name is only needed if we have classes of soft-constraints that should be considered together.
 
-== Optimization examples
-
-  The following examples illustrate the use of basic optimization routines:
-
-     * "Data.SBV.Examples.Optimization.LinearOpt": Simple linear-optimization example.
-     * "Data.SBV.Examples.Optimization.Production": Scheduling machines in a shop
-     * "Data.SBV.Examples.Optimization.VM": Scheduling virtual-machines in a data-center
 -}
 
 {- $modelExtraction
@@ -669,6 +711,31 @@ Floating point numbers are defined by the IEEE-754 standard; and correspond to H
 by Rummer and Wahl: <http://www.philipp.ruemmer.org/publications/smt-fpa.pdf>.
 -}
 
+{- $strings
+Support for characters, strings, and regular expressions (intial version contributed by Joel Burget)
+adds support for QF_S logic, described here: <http://smtlib.cs.uiowa.edu/theories-UnicodeStrings.shtml>
+and here: <http://rise4fun.com/z3/tutorialcontent/sequences>. Note
+that this logic is still not part of official SMTLib (as of March 2018), so it should be considered
+experimental.
+
+See "Data.SBV.Char", "Data.SBV.String", "Data.SBV.RegExp" for related functions.
+-}
+
+{- $lists
+Support for symbolic lists (intial version contributed by Joel Burget)
+adds support for sequence support, described here: <http://rise4fun.com/z3/tutorialcontent/sequences>. Note
+that this logic is still not part of official SMTLib (as of March 2018), so it should be considered
+experimental.
+
+See "Data.SBV.List" for related functions.
+-}
+
+{- $shiftRotate
+Symbolic words (both signed and unsigned) are an instance of Haskell's 'Bits' class, so regular
+bitwise operations are automatically available for them. Shifts and rotates, however, require
+specialized type-signatures since Haskell insists on an 'Int' second argument for them.
+-}
+
 {- $constrainIntro
 A constraint is a means for restricting the input domain of a formula. Here's a simple
 example:
@@ -699,17 +766,19 @@ depends on the context:
     holds. In other words, the constraint says that we only care about
     the input space that satisfies the constraint.
 
-  * In a 'quickCheck' call: The constraint acts as a filter for 'quickCheck';
+  * In a @quickCheck@ call: The constraint acts as a filter for @quickCheck@;
     if the constraint does not hold, then the input value is considered to be irrelevant
     and is skipped. Note that this is similar to 'prove', but is stronger: We do not
     accept a test case to be valid just because the constraints fail on them, although
     semantically the implication does hold. We simply skip that test case as a /bad/
     test vector.
 
-  * In a 'genTest' call: Similar to 'quickCheck' and 'prove': If a constraint
+  * In a 'Data.SBV.Tools.GenTest.genTest' call: Similar to @quickCheck@ and 'prove': If a constraint
     does not hold, the input value is ignored and is not included in the test
     set.
+-}
 
+{- $generalConstraints
 A good use case (in fact the motivating use case) for 'constrain' is attaching a
 constraint to a 'forall' or 'exists' variable at the time of its creation.
 Also, the conjunctive semantics for 'sat' and the implicative
@@ -729,25 +798,12 @@ it's (vacuously) true, since it excludes the entire domain of values, thus makin
 trivial. Hence, this predicate is provable, but is not satisfiable. To make sure the given
 constraints are not vacuous, the functions 'isVacuous' (and 'isVacuousWith') can be used.
 
-Also note that this semantics imply that test case generation ('genTest') and quick-check
-can take arbitrarily long in the presence of constraints, if the random input values generated
+Also note that this semantics imply that test case generation ('Data.SBV.Tools.GenTest.genTest') and
+quick-check can take arbitrarily long in the presence of constraints, if the random input values generated
 rarely satisfy the constraints. (As an extreme case, consider @'constrain' 'false'@.)
+-}
 
-=== Named constraints and unsat cores
-
-Constraints can be given names:
-
-  @ 'namedConstraint' "a is at least 5" $ a .>= 5@
-
-Such constraints are useful when used in conjunction with 'getUnsatCore' function
-where the backend solver can be queried to obtain an unsat core in case the constraints are unsatisfiable.
-This feature is enabled by the following option:
-
-   @ setOption $ ProduceUnsatCores True @
-
-See "Data.SBV.Examples.Misc.UnsatCore" for an example use case.
-
-=== Constraint vacuity
+{- $constraintVacuity
 
 When adding constraints, one has to be careful about
 making sure they are not inconsistent. The function 'isVacuous' can be use for this purpose.
@@ -782,12 +838,25 @@ And the proof is not vacuous:
 
      >>> isVacuous pred'
      False
+-}
 
-=== Checking for vacuity
+{- $namedConstraints
 
-As we discussed SBV does not check that a given constraints is not vacuous. That is, that it can never be satisfied. This is usually
-the right behavior, since checking vacuity can be costly. The functions 'isVacuous' and 'isVacuousWith' should be used
-to explicitly check for constraint vacuity if desired.
+Constraints can be given names:
+
+  @ 'namedConstraint' "a is at least 5" $ a .>= 5@
+
+Similarly, arbitrary term attributes can also be associated:
+
+  @ 'constrainWithAttribute' [(":solver-specific-attribute", "value")] $ a .>= 5@
+
+Note that a 'namedConstraint' is equivalent to a 'constrainWithAttribute' call, setting the `":named"' attribute.
+-}
+
+{- $unsatCores
+Named constraints are useful when used in conjunction with 'Data.SBV.Control.getUnsatCore' function
+where the backend solver can be queried to obtain an unsat core in case the constraints are unsatisfiable.
+See 'Data.SBV.Control.getUnsatCore' for details and "Documentation.SBV.Examples.Queries.UnsatCore" for an example use case.
 -}
 
 {- $uninterpreted
@@ -798,15 +867,15 @@ following example demonstrates:
      data B = B () deriving (Eq, Ord, Show, Read, Data, SymWord, HasKind, SatModel)
   @
 
-(Note that you'll also need to use the language pragmas @DeriveDataTypeable@, @DeriveAnyClass@, and import @Data.Generics@ for the above to work.) 
+(Note that you'll also need to use the language pragmas @DeriveDataTypeable@, @DeriveAnyClass@, and import @Data.Generics@ for the above to work.)
 
-This is all it takes to introduce 'B' as an uninterpreted sort in SBV, which makes the type @SBV B@ automagically become available as the type
-of symbolic values that ranges over 'B' values. Note that the @()@ argument is important to distinguish it from enumerations, which will be
+This is all it takes to introduce @B@ as an uninterpreted sort in SBV, which makes the type @SBV B@ automagically become available as the type
+of symbolic values that ranges over @B@ values. Note that the @()@ argument is important to distinguish it from enumerations, which will be
 translated to proper SMT data-types.
 
 
 Uninterpreted functions over both uninterpreted and regular sorts can be declared using the facilities introduced by
-the 'Uninterpreted' class.
+the 'Data.SBV.Core.Model.Uninterpreted' class.
 -}
 
 {- $enumerations
@@ -857,7 +926,7 @@ which would list all three elements of this domain as satisfying solutions.
 Note that the result is properly typed as @X@ elements; these are not mere strings. So, in a 'getModelAssignment' scenario, the user can recover actual
 elements of the domain and program further with those values as usual.
 
-See "Data.SBV.Examples.Misc.Enumerate" for an extended example on how to use symbolic enumerations.
+See "Documentation.SBV.Examples.Misc.Enumerate" for an extended example on how to use symbolic enumerations.
 -}
 
 {- $noteOnNestedQuantifiers
@@ -865,13 +934,13 @@ See "Data.SBV.Examples.Misc.Enumerate" for an extended example on how to use sym
 
 Note that SBV allows reasoning with quantifiers: Inputs can be existentially or universally quantified. Predicates can be built
 with arbitrary nesting of such quantifiers as well. However, SBV always /assumes/ that the input is in
-prenex-normal form: <https://en.wikipedia.org/wiki/Prenex_normal_form>. That is,
+prenex-normal form: <http://en.wikipedia.org/wiki/Prenex_normal_form>. That is,
 all the input declarations are treated as happening at the beginning of a predicate, followed by the actual formula. Unfortunately,
 the way predicates are written can be misleading at times, since symbolic inputs can be created at arbitrary points; interleaving them
 with other code. The rule is simple, however: All inputs are assumed at the top, in the order declared, regardless of their quantifiers.
 SBV will apply skolemization to get rid of existentials before sending predicates to backend solvers. However, if you do want nested
 quantification, you will manually have to first convert to prenex-normal form (which produces an equisatisfiable but not necessarily
-equivalent formula), and code that explicitly in SBV. See <https://github.com/LeventErkok/sbv/issues/256> for a detailed discussion
+equivalent formula), and code that explicitly in SBV. See <http://github.com/LeventErkok/sbv/issues/256> for a detailed discussion
 of this issue.
 -}
 
@@ -899,4 +968,42 @@ the backend solver supports them. Currently, only Z3 supports pseudo-booleans di
 SBV will translate these to equivalent terms that do not require special functions.
 -}
 
+{- $verbosity
+
+SBV provides various levels of verbosity to aid in debugging, by using the 'SMTConfig' fields:
+
+  * ['verbose'] Print on stdout a shortened account of what is sent/received. This is specifically trimmed to reduce noise
+    and is good for quick debugging. The output is not supposed to be machine-readable.
+  * ['redirectVerbose'] Send the verbose output to a file. Note that you still have to set `verbose=True` for redirection to
+    take effect. Otherwise, the output is the same as what you would see in `verbose`.
+  * ['transcript'] Produce a file that is valid SMTLib2 format, containing everything sent and received. In particular, one can
+    directly feed this file to the SMT-solver outside of the SBV since it is machine-readable. This is good for offline analysis
+    situations, where you want to have a full account of what happened. For instance, it will print time-stamps at every interaction
+    point, so you can see how long each command took.
+-}
+
+{- $observeInternal
+
+The 'observe' command can be used to trace values of arbitrary expressions during a 'sat', 'prove', or perhaps more
+importantly, in a @quickCheck@ call. This is useful for, for instance, recording expected/obtained expressions as a symbolic program is executing.
+
+>>> :{
+prove $ do a1 <- free "i1"
+           a2 <- free "i2"
+           let spec, res :: SWord8
+               spec = a1 + a2
+               res  = ite (a1 .== 12 &&& a2 .== 22)   -- insert a malicious bug!
+                          1
+                          (a1 + a2)
+           observe "Expected" spec
+           observe "Result"   res
+           return $ spec .== res
+:}
+Falsifiable. Counter-example:
+  i1       = 12 :: Word8
+  i2       = 22 :: Word8
+  Expected = 34 :: Word8
+  Result   =  1 :: Word8
+-}
+
 {-# ANN module ("HLint: ignore Use import/export shortcut" :: String) #-}
