@@ -19,7 +19,7 @@
 module Data.SBV.Control.Query (
        send, ask, retrieveResponse
      , CheckSatResult(..), checkSat, checkSatUsing, checkSatAssuming, checkSatAssumingWithUnsatisfiableSet
-     , getUnsatCore, getProof, getInterpolant, getAssignment, getOption, freshVar, freshVar_, push, pop, getAssertionStackDepth
+     , getUnsatCore, getProof, getInterpolant, getAssignment, getOption, freshVar, freshVar_, freshArray, freshArray_, push, pop, getAssertionStackDepth
      , inNewAssertionStack, echo, caseSplit, resetAssertions, exit, getAssertions, getValue, getUninterpretedValue, getModel, getSMTResult
      , getLexicographicOptResults, getIndependentOptResults, getParetoOptResults, getAllSatResult, getUnknownReason
      , SMTOption(..), SMTInfoFlag(..), SMTErrorBehavior(..), SMTReasonUnknown(..), SMTInfoResponse(..), getInfo
@@ -33,8 +33,13 @@ module Data.SBV.Control.Query (
 import Control.Monad            (unless, when, zipWithM)
 import Control.Monad.State.Lazy (get)
 
-import Data.IORef    (readIORef)
+import Data.IORef (readIORef)
 
+import qualified Data.Map.Strict    as M
+import qualified Data.IntMap.Strict as IM
+
+
+import Data.Char     (toLower)
 import Data.List     (unzip3, intercalate, nubBy, sortBy)
 import Data.Maybe    (listToMaybe, catMaybes)
 import Data.Function (on)
@@ -107,8 +112,6 @@ getInfo flag = do
 
         isAllStat = isAllStatistics flag
 
-        render = serialize True
-
         grabAllStat k v = (render k, render v)
 
         -- we're trying to do our best to get key-value pairs here, but this
@@ -131,13 +134,25 @@ getInfo flag = do
                  EApp [ECon ":error-behavior", ECon "immediate-exit"]      -> return $ Resp_Error ErrorImmediateExit
                  EApp [ECon ":error-behavior", ECon "continued-execution"] -> return $ Resp_Error ErrorContinuedExecution
                  EApp (ECon ":name" : o)                                   -> return $ Resp_Name (render (EApp o))
-                 EApp [ECon ":reason-unknown", ECon "memout"]              -> return $ Resp_ReasonUnknown UnknownMemOut
-                 EApp [ECon ":reason-unknown", ECon "incomplete"]          -> return $ Resp_ReasonUnknown UnknownIncomplete
-                 EApp (ECon ":reason-unknown" : o)                         -> return $ Resp_ReasonUnknown (UnknownOther (render (EApp o)))
+                 EApp (ECon ":reason-unknown" : o)                         -> return $ Resp_ReasonUnknown (unk o)
                  EApp (ECon ":version" : o)                                -> return $ Resp_Version (render (EApp o))
                  EApp (ECon s : o)                                         -> return $ Resp_InfoKeyword s (map render o)
                  _                                                         -> bad r Nothing
 
+  where render = serialize True
+
+        unk [ECon s] | Just d <- getUR s = d
+        unk o                            = UnknownOther (render (EApp o))
+
+        getUR s = map toLower (unQuote s) `lookup` [(map toLower k, d) | (k, d) <- unknownReasons]
+
+        -- As specified in Section 4.1 of the SMTLib document. Note that we're adding the
+        -- extra timeout as it is useful in this context.
+        unknownReasons = [ ("memout",     UnknownMemOut)
+                         , ("incomplete", UnknownIncomplete)
+                         , ("timeout",    UnknownTimeOut)
+                         ]
+
 -- | Retrieve the value of an 'SMTOption.' The curious function argument is on purpose here,
 -- simply pass the constructor name. Example: the call @'getOption' 'ProduceUnsatCores'@ will return
 -- either @Nothing@ or @Just (ProduceUnsatCores True)@ or @Just (ProduceUnsatCores False)@.
@@ -183,25 +198,22 @@ getOption f = case f undefined of
         stringList c e _ = return $ Just $ c $ stringsOf e
 
 -- | Get the reason unknown. Only internally used.
-getUnknownReason :: Query String
+getUnknownReason :: Query SMTReasonUnknown
 getUnknownReason = do ru <- getInfo ReasonUnknown
                       case ru of
-                        Resp_Unsupported     -> return "No reason provided."
-                        Resp_ReasonUnknown r -> return $ case r of
-                                                           UnknownMemOut       -> "Out of memory."
-                                                           UnknownIncomplete   -> "Incomplete."
-                                                           UnknownOther      s -> s
+                        Resp_Unsupported     -> return $ UnknownOther "Solver responded: Unsupported."
+                        Resp_ReasonUnknown r -> return r
                         -- Shouldn't happen, but just in case:
-                        _                    -> return $ "Unexpected reason value received: " ++ show ru
+                        _                    -> error $ "Unexpected reason value received: " ++ show ru
 
 -- | Issue check-sat and get an SMT Result out.
 getSMTResult :: Query SMTResult
 getSMTResult = do cfg <- getConfig
                   cs  <- checkSat
                   case cs of
-                    Unsat -> return $ Unsatisfiable cfg
-                    Sat   -> Satisfiable cfg <$> getModel
-                    Unk   -> Unknown     cfg <$> getUnknownReason
+                    Unsat -> Unsatisfiable cfg <$> getUnsatCoreIfRequested
+                    Sat   -> Satisfiable   cfg <$> getModel
+                    Unk   -> Unknown       cfg <$> getUnknownReason
 
 -- | Classify a model based on whether it has unbound objectives or not.
 classifyModel :: SMTConfig -> SMTModel -> SMTResult
@@ -214,7 +226,7 @@ getLexicographicOptResults :: Query SMTResult
 getLexicographicOptResults = do cfg <- getConfig
                                 cs  <- checkSat
                                 case cs of
-                                  Unsat -> return $ Unsatisfiable cfg
+                                  Unsat -> Unsatisfiable cfg <$> getUnsatCoreIfRequested
                                   Sat   -> classifyModel cfg <$> getModelWithObjectives
                                   Unk   -> Unknown       cfg <$> getUnknownReason
    where getModelWithObjectives = do objectiveValues <- getObjectiveValues
@@ -227,7 +239,7 @@ getIndependentOptResults objNames = do cfg <- getConfig
                                        cs  <- checkSat
 
                                        case cs of
-                                         Unsat -> return [(nm, Unsatisfiable cfg) | nm <- objNames]
+                                         Unsat -> getUnsatCoreIfRequested >>= \mbUC -> return [(nm, Unsatisfiable cfg mbUC) | nm <- objNames]
                                          Sat   -> continue (classifyModel cfg)
                                          Unk   -> do ur <- Unknown cfg <$> getUnknownReason
                                                      return [(nm, ur) | nm <- objNames]
@@ -251,7 +263,7 @@ getParetoOptResults mbN      = do cfg <- getConfig
                                     Unsat -> return (False, [])
                                     Sat   -> continue (classifyModel cfg)
                                     Unk   -> do ur <- getUnknownReason
-                                                return (False, [ProofError cfg [ur]])
+                                                return (False, [ProofError cfg [show ur]])
 
   where continue classify = do m <- getModel
                                (limReached, fronts) <- getParetoFronts (subtract 1 <$> mbN) [m]
@@ -276,9 +288,10 @@ getModel = getModelAtIndex Nothing
 getModelAtIndex :: Maybe Int -> Query SMTModel
 getModelAtIndex mbi = do
              State{runMode} <- get
-             cfg  <- getConfig
-             inps <- getQuantifiedInputs
-             rm   <- io $ readIORef runMode
+             cfg   <- getConfig
+             inps  <- getQuantifiedInputs
+             obsvs <- getObservables
+             rm    <- io $ readIORef runMode
              let vars :: [NamedSymVar]
                  vars = case rm of
                           m@CodeGen         -> error $ "SBV.getModel: Model is not available in mode: " ++ show m
@@ -287,10 +300,17 @@ getModelAtIndex mbi = do
                                                let allModelInputs = if isSAT then takeWhile ((/= ALL) . fst) inps
                                                                              else takeWhile ((== ALL) . fst) inps
 
+                                                   -- are we inside a quantifier
+                                                   insideQuantifier = length allModelInputs < length inps
+
+                                                   -- observables are only meaningful if we're not in a quantified context
+                                                   allPrefixObservables | insideQuantifier = []
+                                                                        | True             = [(EX, (sw, nm)) | (nm, sw) <- obsvs]
+
                                                    sortByNodeId :: [NamedSymVar] -> [NamedSymVar]
                                                    sortByNodeId = sortBy (compare `on` (\(SW _ n, _) -> n))
 
-                                               in sortByNodeId [nv | (_, nv@(_, n)) <- allModelInputs, not (isNonModelVar cfg n)]
+                                               in sortByNodeId [nv | (_, nv@(_, n)) <- allModelInputs ++ allPrefixObservables, not (isNonModelVar cfg n)]
 
              assocs <- mapM (\(sw, n) -> (n, ) <$> getValueCW mbi sw) vars
 
@@ -309,8 +329,7 @@ getObjectiveValues = do let cmd = "(get-objectives)"
 
                         inputs <- map snd <$> getQuantifiedInputs
 
-                        parse r bad $ \case EApp (ECon "objectives" : es) -> do mbVals <- mapM (getObjValue (bad r) inputs) es
-                                                                                return $ catMaybes mbVals
+                        parse r bad $ \case EApp (ECon "objectives" : es) -> catMaybes <$> mapM (getObjValue (bad r) inputs) es
                                             _                             -> bad r Nothing
 
   where -- | Parse an objective value out.
@@ -356,24 +375,29 @@ getObjectiveValues = do let cmd = "(get-objectives)"
                         simplify (EApp xs)                  = EApp (map simplify xs)
                         simplify e                          = e
 
--- | Check for satisfiability, under the given conditions. Similar to 'checkSat' except it allows making
+-- | Check for satisfiability, under the given conditions. Similar to 'Data.SBV.Control.checkSat' except it allows making
 -- further assumptions as captured by the first argument of booleans. (Also see 'checkSatAssumingWithUnsatisfiableSet'
 -- for a variant that returns the subset of the given assumptions that led to the 'Unsat' conclusion.)
 checkSatAssuming :: [SBool] -> Query CheckSatResult
 checkSatAssuming sBools = fst <$> checkSatAssumingHelper False sBools
 
 -- | Check for satisfiability, under the given conditions. Returns the unsatisfiable
--- set of assumptions. Similar to 'checkSat' except it allows making further assumptions
+-- set of assumptions. Similar to 'Data.SBV.Control.checkSat' except it allows making further assumptions
 -- as captured by the first argument of booleans. If the result is 'Unsat', the user will
 -- also receive a subset of the given assumptions that led to the 'Unsat' conclusion. Note
 -- that while this set will be a subset of the inputs, it is not necessarily guaranteed to be minimal.
 --
 -- You must have arranged for the production of unsat assumptions
--- first (/via/ @'setOption' $ 'ProduceUnsatAssumptions' 'True'@)
+-- first via
+--
+-- @
+--     'setOption' $ 'ProduceUnsatAssumptions' 'True'
+-- @
+--
 -- for this call to not error out!
 --
 -- Usage note: 'getUnsatCore' is usually easier to use than 'checkSatAssumingWithUnsatisfiableSet', as it
--- allows the use of named assertions, as obtained by 'namedAssert'. If 'getUnsatCore'
+-- allows the use of named assertions, as obtained by 'namedConstraint'. If 'getUnsatCore'
 -- fills your needs, you should definitely prefer it over 'checkSatAssumingWithUnsatisfiableSet'.
 checkSatAssumingWithUnsatisfiableSet :: [SBool] -> Query (CheckSatResult, Maybe [SBool])
 checkSatAssumingWithUnsatisfiableSet = checkSatAssumingHelper True
@@ -431,6 +455,37 @@ checkSatAssumingHelper getAssumptions sBools = do
 getAssertionStackDepth :: Query Int
 getAssertionStackDepth = queryAssertionStackDepth <$> getQueryState
 
+-- | Upon a pop, we need to restore all arrays and tables. See: http://github.com/LeventErkok/sbv/issues/374
+restoreTablesAndArrays :: Query ()
+restoreTablesAndArrays = do st <- get
+                            qs <- getQueryState
+
+                            case queryTblArrPreserveIndex qs of
+                              Nothing       -> return ()
+                              Just (tc, ac) -> do tCount <- M.size  <$> (io . readIORef) (rtblMap   st)
+                                                  aCount <- IM.size <$> (io . readIORef) (rArrayMap st)
+
+                                                  let tInits = [ "table"  ++ show i ++ "_initializer" | i <- [tc .. tCount - 1]]
+                                                      aInits = [ "array_" ++ show i ++ "_initializer" | i <- [ac .. aCount - 1]]
+                                                      inits  = tInits ++ aInits
+
+                                                  case inits of
+                                                    []  -> return ()   -- Nothing to do
+                                                    [x] -> send True $ "(assert " ++ x ++ ")"
+                                                    xs  -> send True $ "(assert (and " ++ unwords xs ++ "))"
+
+-- | Upon a push, record the cut-off point for table and array restoration, if we haven't already
+recordTablesAndArrayCutOff :: Query ()
+recordTablesAndArrayCutOff = do st <- get
+                                qs <- getQueryState
+
+                                case queryTblArrPreserveIndex qs of
+                                  Just _  -> return () -- already recorded, nothing to do
+                                  Nothing -> do tCount <- M.size  <$> (io . readIORef) (rtblMap   st)
+                                                aCount <- IM.size <$> (io . readIORef) (rArrayMap st)
+
+                                                modifyQueryState $ \s -> s {queryTblArrPreserveIndex = Just (tCount, aCount)}
+
 -- | Run the query in a new assertion stack. That is, we push the context, run the query
 -- commands, and pop it back.
 inNewAssertionStack :: Query a -> Query a
@@ -445,6 +500,7 @@ push i
  | i <= 0 = error $ "Data.SBV: push requires a strictly positive level argument, received: " ++ show i
  | True   = do depth <- getAssertionStackDepth
                send True $ "(push " ++ show i ++ ")"
+               recordTablesAndArrayCutOff
                modifyQueryState $ \s -> s{queryAssertionStackDepth = depth + i}
 
 -- | Pop the context, exiting a new one. Pops multiple levels if /n/ > 1. It's an error to pop levels that don't exist.
@@ -463,6 +519,7 @@ pop i
                                                   , "*** Request this as a feature for the underlying solver!"
                                                   ]
                              else do send True $ "(pop " ++ show i ++ ")"
+                                     restoreTablesAndArrays
                                      modifyQueryState $ \s -> s{queryAssertionStackDepth = depth - i}
    where shl 1 = "one level"
          shl n = show n ++ " levels"
@@ -471,7 +528,7 @@ pop i
 -- the conditions lead to a satisfiable result, returns @Just@ that result. If none of them
 -- do, returns @Nothing@. Note that we automatically generate a coverage case and search
 -- for it automatically as well. In that latter case, the string returned will be "Coverage".
--- The first argument controls printing progress messages  See "Data.SBV.Examples.Queries.CaseSplit"
+-- The first argument controls printing progress messages  See "Documentation.SBV.Examples.Queries.CaseSplit"
 -- for an example use case.
 caseSplit :: Bool -> [(String, SBool)] -> Query (Maybe (String, SMTResult))
 caseSplit printCases cases = do cfg <- getConfig
@@ -497,9 +554,10 @@ caseSplit printCases cases = do cfg <- getConfig
                                               return $ Just (n, res)
 
 -- | Reset the solver, by forgetting all the assertions. However, bindings are kept as is,
--- as opposed to 'reset'. Use this variant to clean-up the solver state while leaving the bindings
--- intact. Pops all assertion levels. Declarations and definitions resulting from the 'setLogic'
--- command are unaffected. Note that SBV implicitly uses global-declarations, so bindings will remain intact.
+-- as opposed to a full reset of the solver. Use this variant to clean-up the solver
+-- state while leaving the bindings intact. Pops all assertion levels. Declarations and
+-- definitions resulting from the 'Data.SBV.setLogic' command are unaffected. Note that SBV
+-- implicitly uses global-declarations, so bindings will remain intact.
 resetAssertions :: Query ()
 resetAssertions = do send True "(reset-assertions)"
                      modifyQueryState $ \s -> s{queryAssertionStackDepth = 0}
@@ -526,8 +584,25 @@ exit = do send True "(exit)"
           modifyQueryState $ \s -> s{queryAssertionStackDepth = 0}
 
 -- | Retrieve the unsat-core. Note you must have arranged for
--- unsat cores to be produced first (/via/ @'setOption' $ 'ProduceUnsatCores' 'True'@)
+-- unsat cores to be produced first via
+--
+-- @
+--     'setOption' $ 'ProduceUnsatCores' 'True'
+-- @
+--
 -- for this call to not error out!
+--
+-- NB. There is no notion of a minimal unsat-core, in case unsatisfiability can be derived
+-- in multiple ways. Furthermore, Z3 does not guarantee that the generated unsat
+-- core does not have any redundant assertions either, as doing so can incur a performance penalty.
+-- (There might be assertions in the set that is not needed.) To ensure all the assertions
+-- in the core are relevant, use:
+--
+-- @
+--     'setOption' $ 'OptionKeyword' ":smt.core.minimize" ["true"]
+-- @
+--
+-- Note that this only works with Z3.
 getUnsatCore :: Query [String]
 getUnsatCore = do
         let cmd = "(get-unsat-core)"
@@ -538,6 +613,15 @@ getUnsatCore = do
                                   , ""
                                   , "so the solver will be ready to compute unsat cores,"
                                   , "and that there is a model by first issuing a 'checkSat' call."
+                                  , ""
+                                  , "If using z3, you might also optionally want to set:"
+                                  , ""
+                                  , "       setOption $ OptionKeyword \":smt.core.minimize\" [\"true\"]"
+                                  , ""
+                                  , "to make sure the unsat core doesn't have irrelevant entries,"
+                                  , "though this might incur a performance penalty."
+
+
                                   ]
 
         r <- ask cmd
@@ -546,8 +630,21 @@ getUnsatCore = do
            EApp es | Just xs <- mapM fromECon es -> return $ map unBar xs
            _                                     -> bad r Nothing
 
+-- | Retrieve the unsat core if it was asked for in the configuration
+getUnsatCoreIfRequested :: Query (Maybe [String])
+getUnsatCoreIfRequested = do
+        cfg <- getConfig
+        if or [b | ProduceUnsatCores b <- solverSetOptions cfg]
+           then Just <$> getUnsatCore
+           else return Nothing
+
 -- | Retrieve the proof. Note you must have arranged for
--- proofs to be produced first (/via/ @'setOption' $ 'ProduceProofs' 'True'@)
+-- proofs to be produced first via
+--
+-- @
+--     'setOption' $ 'ProduceProofs' 'True'
+-- @
+--
 -- for this call to not error out!
 --
 -- A proof is simply a 'String', as returned by the solver. In the future, SBV might
@@ -572,13 +669,18 @@ getProof = do
         -- result of parsing is ignored.
         parse r bad $ \_ -> return r
 
--- | Retrieve interpolants after an 'Unsat' result is obtained. Note you must have arranged for
--- interpolants to be produced first (/via/ @'setOption' $ 'ProduceInterpolants' 'True'@)
+-- | Retrieve an interpolant after an 'Unsat' result is obtained. Note you must have arranged for
+-- interpolants to be produced first via
+--
+-- @
+--     'setOption' $ 'ProduceInterpolants' 'True'
+-- @
+--
 -- for this call to not error out!
 --
--- To get an interpolant for a pair of formulas @A@ and @B@, use a 'namedConstraint' to attach
--- names to @A@ and @B@. Then call 'getInterpolant' @[\"A\", \"B\"]@, assuming those are the names
--- you gave to the formulas.
+-- To get an interpolant for a pair of formulas @A@ and @B@, use a 'constrainWithAttribute' call to attach
+-- interplation groups to @A@ and @B@. Then call 'getInterpolant' @[\"A\"]@, assuming those are the names
+-- you gave to the formulas in the @A@ group.
 --
 -- An interpolant for @A@ and @B@ is a formula @I@ such that:
 --
@@ -592,38 +694,36 @@ getProof = do
 -- be satisfied at the same time. Furthermore, @I@ will have only the symbols that are common
 -- to @A@ and @B@.
 --
--- Interpolants generalize to sequences: If you pass more than two formulas, then you will get
--- a sequence of interpolants. In general, for @N@ formulas that are not satisfiable together, you will be
--- returned @N-1@ interpolants. If formulas are @A1 .. An@, then interpolants will be @I1 .. I(N-1)@, such
--- that @A1 ==> I1@, @A2 /\\ I1 ==> I2@, @A3 /\\ I2 ==> I3@, ..., and finally @AN ===> not I(N-1)@.
---
--- Currently, SBV only returns simple and sequence interpolants, and does not support tree-interpolants.
--- If you need these, please get in touch. Furthermore, the result will be a list of mere strings representing the
--- interpolating formulas, as opposed to a more structured type. Please get in touch if you use this function and can
--- suggest a better API.
-getInterpolant :: [String] -> Query [String]
+-- N.B. As of Z3 version 4.8.0; Z3 no longer supports interpolants. Use the MathSAT backend for extracting
+-- interpolants. See "Documentation.SBV.Examples.Queries.Interpolants" for an example.
+getInterpolant :: [String] -> Query String
 getInterpolant fs
-  | length fs < 2
-  = error $ "SBV.getInterpolant requires at least two named constraints, received: " ++ show fs
+  | null fs
+  = error "SBV.getInterpolant requires at least one marked constraint, received none!"
   | True
   = do let bar s = '|' : s ++ "|"
-           cmd = "(get-interpolant " ++ unwords (map bar fs) ++ ")"
+           cmd = "(get-interpolant (" ++ unwords (map bar fs) ++ "))"
            bad = unexpected "getInterpolant" cmd "a get-interpolant response"
                           $ Just [ "Make sure you use:"
                                  , ""
                                  , "       setOption $ ProduceInterpolants True"
                                  , ""
                                  , "to make sure the solver is ready for producing interpolants,"
-                                 , "and that you have named the formulas with calls to 'namedConstraint'."
+                                 , "and that you have used the proper attributes using the"
+                                 , "constrainWithAttribute function."
                                  ]
 
        r <- ask cmd
 
-       parse r bad $ \case EApp (ECon "interpolants" : es) -> return $ map (serialize False) es
-                           _                               -> bad r Nothing
+       parse r bad $ \e -> return $ serialize False e
 
 -- | Retrieve assertions. Note you must have arranged for
--- assertions to be available first (/via/ @'setOption' $ 'ProduceAssertions' 'True'@)
+-- assertions to be available first via
+--
+-- @
+--     'setOption' $ 'ProduceAssertions' 'True'
+-- @
+--
 -- for this call to not error out!
 --
 -- Note that the set of assertions returned is merely a list of strings, just like the
@@ -651,6 +751,14 @@ getAssertions = do
 
 -- | Retrieve the assignment. This is a lightweight version of 'getValue', where the
 -- solver returns the truth value for all named subterms of type 'Bool'.
+--
+-- You must have first arranged for assignments to be produced via
+--
+-- @
+--     'setOption' $ 'ProduceAssignments' 'True'
+-- @
+--
+-- for this call to not error out!
 getAssignment :: Query [(String, Bool)]
 getAssignment = do
         let cmd = "(get-assignment)"
