@@ -9,20 +9,22 @@
 -- Query related utils.
 -----------------------------------------------------------------------------
 
-{-# LANGUAGE    BangPatterns          #-}
-{-# LANGUAGE    DefaultSignatures     #-}
-{-# LANGUAGE    LambdaCase            #-}
-{-# LANGUAGE    NamedFieldPuns        #-}
-{-# LANGUAGE    ScopedTypeVariables   #-}
-{-# LANGUAGE    TupleSections         #-}
-{-# OPTIONS_GHC -fno-warn-orphans  #-}
+{-# LANGUAGE    BangPatterns         #-}
+{-# LANGUAGE    DefaultSignatures    #-}
+{-# LANGUAGE    LambdaCase           #-}
+{-# LANGUAGE    NamedFieldPuns       #-}
+{-# LANGUAGE    ScopedTypeVariables  #-}
+{-# LANGUAGE    TupleSections        #-}
+{-# LANGUAGE    TypeSynonymInstances #-}
+{-# LANGUAGE    FlexibleInstances    #-}
+{-# OPTIONS_GHC -fno-warn-orphans    #-}
 
 module Data.SBV.Control.Utils (
        io
      , ask, send, getValue, getUninterpretedValue, getValueCW, getUnsatAssumptions, SMTValue(..)
-     , getQueryState, modifyQueryState, getConfig, getObjectives, getSBVAssertions, getSBVPgm, getQuantifiedInputs
+     , getQueryState, modifyQueryState, getConfig, getObjectives, getSBVAssertions, getSBVPgm, getQuantifiedInputs, getObservables
      , checkSat, checkSatUsing, getAllSatResult
-     , inNewContext, freshVar, freshVar_
+     , inNewContext, freshVar, freshVar_, freshArray, freshArray_
      , parse
      , unexpected
      , timeout
@@ -30,23 +32,30 @@ module Data.SBV.Control.Utils (
      , retrieveResponse
      , recoverKindedValue
      , runProofOn
+     , executeQuery
      ) where
 
-import Data.List  (sortBy, elemIndex, partition, groupBy, tails)
+import Data.Maybe (isJust)
+import Data.List  (sortBy, sortOn, elemIndex, partition, groupBy, tails, intercalate)
 
-import Data.Char     (isPunctuation, isSpace)
-import Data.Ord      (comparing)
+import Data.Char     (isPunctuation, isSpace, chr, ord)
 import Data.Function (on)
 
+import Data.Typeable (Typeable)
+
 import Data.Int
 import Data.Word
 
-import qualified Data.Map    as Map
-import qualified Data.IntMap as IMap
+import qualified Data.Map.Strict    as Map
+import qualified Data.IntMap.Strict as IMap
+
+import qualified Control.Monad.Reader as R (ask)
 
 import Control.Monad            (unless)
 import Control.Monad.State.Lazy (get, liftIO)
 
+import Control.Monad.State      (evalStateT)
+
 import Data.IORef (readIORef, writeIORef)
 
 import Data.Time (getZonedTime)
@@ -55,17 +64,25 @@ import Data.SBV.Core.Data     ( SW(..), CW(..), SBV, AlgReal, sbvToSW, kindOf, K
                               , HasKind(..), mkConstCW, CWVal(..), SMTResult(..)
                               , NamedSymVar, SMTConfig(..), Query, SMTModel(..)
                               , QueryState(..), SVal(..), Quantifier(..), cache
-                              , newExpr, SBVExpr(..), Op(..), FPOp(..), SBV(..)
+                              , newExpr, SBVExpr(..), Op(..), FPOp(..), SBV(..), SymArray(..)
                               , SolverContext(..), SBool, Objective(..), SolverCapabilities(..), capabilities
-                              , Result(..), SMTProblem(..), trueSW, SymWord(..), SBVPgm(..), SMTSolver(..)
+                              , Result(..), SMTProblem(..), trueSW, SymWord(..), SBVPgm(..), SMTSolver(..), SBVRunMode(..)
+                              )
+
+import Data.SBV.Core.Symbolic ( IncState(..), withNewIncState, State(..), svToSW, Symbolic
+                              , QueryContext(..)
+                              , registerLabel, svMkSymVar
+                              , isSafetyCheckingIStage, isSetupIStage, isRunIStage, IStage(..), Query(..)
+                              , extractSymbolicSimulationState
                               )
-import Data.SBV.Core.Symbolic (IncState(..), withNewIncState, State(..), svToSW, registerLabel, svMkSymVar)
 
 import Data.SBV.Core.AlgReals   (mergeAlgReals)
 import Data.SBV.Core.Operations (svNot, svNotEqual, svOr)
 
 import Data.SBV.SMT.SMTLib  (toIncSMTLib, toSMTLib)
-import Data.SBV.SMT.Utils   (showTimeoutValue, annotateWithName, alignPlain, debug, mergeSExpr, SMTException(..))
+import Data.SBV.SMT.Utils   (showTimeoutValue, addAnnotations, alignPlain, debug, mergeSExpr, SBVException(..))
+
+import Data.SBV.Utils.Lib (qfsToString, isKString)
 
 import Data.SBV.Utils.SExpr
 import Data.SBV.Control.Types
@@ -76,10 +93,14 @@ import qualified Control.Exception as C
 
 import GHC.Stack
 
+import Unsafe.Coerce (unsafeCoerce) -- Only used safely!
+
 -- | 'Query' as a 'SolverContext'.
 instance SolverContext Query where
-   constrain          = addQueryConstraint Nothing
-   namedConstraint nm = addQueryConstraint (Just nm)
+   constrain                 = addQueryConstraint False []
+   softConstrain             = addQueryConstraint True  []
+   namedConstraint nm        = addQueryConstraint False [(":named", nm)]
+   constrainWithAttribute    = addQueryConstraint False
 
    setOption o
      | isStartModeOption o = error $ unlines [ ""
@@ -88,15 +109,14 @@ instance SolverContext Query where
                                              ]
      | True                = send True $ setSMTOption o
 
--- | Adding a constraint, possibly named. Only used internally.
+-- | Adding a constraint, possibly with attributes and possibly soft. Only used internally.
 -- Use 'constrain' and 'namedConstraint' from user programs.
-addQueryConstraint :: Maybe String -> SBool -> Query ()
-addQueryConstraint mbNm b = do sw <- inNewContext (\st -> do maybe (return ()) (registerLabel st) mbNm
-                                                             sbvToSW st b)
-                               send True $ "(assert " ++ mkNamed mbNm (show sw)  ++ ")"
-   where mkNamed Nothing   s = s
-         mkNamed (Just nm) s = annotateWithName nm s
-
+addQueryConstraint :: Bool -> [(String, String)] -> SBool -> Query ()
+addQueryConstraint isSoft atts b = do sw <- inNewContext (\st -> do mapM_ (registerLabel "Constraint" st) [nm | (":named", nm) <- atts]
+                                                                    sbvToSW st b)
+                                      send True $ "(" ++ asrt ++ " " ++ addAnnotations atts (show sw)  ++ ")"
+   where asrt | isSoft = "assert-soft"
+              | True   = "assert"
 
 -- | Get the current configuration
 getConfig :: Query SMTConfig
@@ -112,7 +132,7 @@ getSBVPgm :: Query SBVPgm
 getSBVPgm = do State{spgm} <- get
                io $ readIORef spgm
 
--- | Get the assertions put in via 'sAssert'
+-- | Get the assertions put in via 'Data.SBV.sAssert'
 getSBVAssertions :: Query [(String, Maybe CallStack, SW)]
 getSBVAssertions = do State{rAsserts} <- get
                       io $ reverse <$> readIORef rAsserts
@@ -122,20 +142,20 @@ io :: IO a -> Query a
 io = liftIO
 
 -- | Sync-up the external solver with new context we have generated
-syncUpSolver :: IncState -> Query ()
-syncUpSolver is = do
+syncUpSolver :: Bool -> IncState -> Query ()
+syncUpSolver afterAPush is = do
         cfg <- getConfig
         ls  <- io $ do let swap  (a, b)        = (b, a)
-                           swapc ((_, a), b)   = (b, a)
                            cmp   (a, _) (b, _) = a `compare` b
                            arrange (i, (at, rt, es)) = ((i, at, rt), es)
                        inps  <- reverse <$> readIORef (rNewInps is)
                        ks    <- readIORef (rNewKinds is)
-                       cnsts <- sortBy cmp . map swapc . Map.toList <$> readIORef (rNewConsts is)
+                       cnsts <- sortBy cmp . map swap . Map.toList <$> readIORef (rNewConsts is)
                        arrs  <- IMap.toAscList <$> readIORef (rNewArrs is)
-                       tbls  <- (map arrange . sortBy cmp . map swap . Map.toList) <$> readIORef (rNewTbls is)
+                       tbls  <- map arrange . sortBy cmp . map swap . Map.toList <$> readIORef (rNewTbls is)
+                       uis   <- Map.toAscList <$> readIORef (rNewUIs is)
                        as    <- readIORef (rNewAsgns is)
-                       return $ toIncSMTLib cfg inps ks cnsts arrs tbls as cfg
+                       return $ toIncSMTLib afterAPush cfg inps ks cnsts arrs tbls uis as cfg
         mapM_ (send True) $ mergeSExpr ls
 
 -- | Retrieve the query context
@@ -165,7 +185,11 @@ modifyQueryState f = do state <- get
 inNewContext :: (State -> IO a) -> Query a
 inNewContext act = do st <- get
                       (is, r) <- io $ withNewIncState st act
-                      syncUpSolver is
+                      mbQS <- io . readIORef . queryState $ st
+                      let afterAPush = case mbQS of
+                                         Nothing -> False
+                                         Just qs -> isJust (queryTblArrPreserveIndex qs)
+                      syncUpSolver afterAPush is
                       return r
 
 -- | Similar to 'freshVar', except creates unnamed variable.
@@ -174,7 +198,7 @@ freshVar_ = inNewContext $ fmap SBV . svMkSymVar (Just EX) k Nothing
   where k = kindOf (undefined :: a)
 
 -- | Create a fresh variable in query mode. You should prefer
--- creating input variables using 'sBool', 'sInt32', etc., which act
+-- creating input variables using 'Data.SBV.sBool', 'Data.SBV.sInt32', etc., which act
 -- as primary inputs to the model and can be existential or universal.
 -- Use 'freshVar' only in query mode for anonymous temporary variables.
 -- Such variables are always existential. Note that 'freshVar' should hardly be
@@ -184,6 +208,21 @@ freshVar :: forall a. SymWord a => String -> Query (SBV a)
 freshVar nm = inNewContext $ fmap SBV . svMkSymVar (Just EX) k (Just nm)
   where k = kindOf (undefined :: a)
 
+-- | Similar to 'freshArray', except creates unnamed array.
+freshArray_ :: (SymArray array, HasKind a, HasKind b) => Maybe (SBV b) -> Query (array a b)
+freshArray_ = mkFreshArray Nothing
+
+-- | Create a fresh array in query mode. Again, you should prefer
+-- creating arrays before the queries start using 'newArray', but this
+-- method can come in handy in occasional cases where you need a new array
+-- after you start the query based interaction.
+freshArray :: (SymArray array, HasKind a, HasKind b) => String -> Maybe (SBV b) -> Query (array a b)
+freshArray nm = mkFreshArray (Just nm)
+
+-- | Creating arrays, internal use only.
+mkFreshArray :: (SymArray array, HasKind a, HasKind b) => Maybe String -> Maybe (SBV b) -> Query (array a b)
+mkFreshArray mbNm mbVal = inNewContext $ newArrayInState mbNm mbVal
+
 -- | If 'verbose' is 'True', print the message, useful for debugging messages
 -- in custom queries. Note that 'redirectVerbose' will be respected: If a
 -- file redirection is given, the output will go to the file.
@@ -203,6 +242,30 @@ ask s = do QueryState{queryAsk, queryTimeOutValue} <- getQueryState
 
            return r
 
+-- | Send a string to the solver, and return the response. Except, if the response
+-- is one of the "ignore" ones, keep querying.
+askIgnoring :: String -> [String] -> Query String
+askIgnoring s ignoreList = do
+
+           QueryState{queryAsk, queryRetrieveResponse, queryTimeOutValue} <- getQueryState
+
+           case queryTimeOutValue of
+             Nothing -> queryDebug ["[SEND] " `alignPlain` s]
+             Just i  -> queryDebug ["[SEND, TimeOut: " ++ showTimeoutValue i ++ "] " `alignPlain` s]
+           r <- io $ queryAsk queryTimeOutValue s
+           queryDebug ["[RECV] " `alignPlain` r]
+
+           let loop currentResponse
+                 | currentResponse `notElem` ignoreList
+                 = return currentResponse
+                 | True
+                 = do queryDebug ["[WARN] Previous response is explicitly ignored, beware!"]
+                      newResponse <- io $ queryRetrieveResponse queryTimeOutValue
+                      queryDebug ["[RECV] " `alignPlain` newResponse]
+                      loop newResponse
+
+           loop r
+
 -- | Send a string to the solver. If the first argument is 'True', we will require
 -- a "success" response as well. Otherwise, we'll fire and forget.
 send :: Bool -> String -> Query ()
@@ -306,6 +369,34 @@ instance SMTValue AlgReal where
    sexprToVal (ENum (v, _)) = Just (fromIntegral v)
    sexprToVal _             = Nothing
 
+instance SMTValue Char where
+   sexprToVal (ENum (i, _)) = Just (chr (fromIntegral i))
+   sexprToVal _             = Nothing
+
+instance (SMTValue a, Typeable a) => SMTValue [a] where
+   -- NB. The conflation of String/[Char] forces us to have this bastard case here
+   -- with unsafeCoerce to cast back to a regular string. This is unfortunate,
+   -- and the ice is thin here. But it works, and is much better than a plethora
+   -- of overlapping instances. Sigh.
+   sexprToVal (ECon s)
+    | isKString (undefined :: [a]) && length s >= 2 && head s == '"' && last s == '"'
+    = Just $ map unsafeCoerce s'
+    | True
+    = Just $ map (unsafeCoerce . c2w8) s'
+    where s' = qfsToString (tail (init s))
+          c2w8  :: Char -> Word8
+          c2w8 = fromIntegral . ord
+
+   -- Otherwise we have a good old sequence, just parse it simply:
+   sexprToVal (EApp [ECon "seq.++", l, r])            = do l' <- sexprToVal l
+                                                           r' <- sexprToVal r
+                                                           return $ l' ++ r'
+   sexprToVal (EApp [ECon "seq.unit", a])             = do a' <- sexprToVal a
+                                                           return [a']
+   sexprToVal (EApp [ECon "as", ECon "seq.empty", _]) = return []
+
+   sexprToVal _                                       = Nothing
+
 -- | Get the value of a term.
 getValue :: SMTValue a => SBV a -> Query a
 getValue s = do sw <- inNewContext (`sbvToSW` s)
@@ -331,7 +422,7 @@ getUninterpretedValue s =
                                      r <- ask cmd
 
                                      parse r bad $ \case EApp [EApp [ECon o,  ECon v]] | o == show sw -> return v
-                                                         _                                             -> bad r Nothing
+                                                         _                                            -> bad r Nothing
 
           k                    -> error $ unlines [""
                                                   , "*** SBV.getUninterpretedValue: Called on an 'interpreted' kind"
@@ -367,16 +458,50 @@ getValueCWHelper mbi s = do
 recoverKindedValue :: Kind -> SExpr -> Maybe CW
 recoverKindedValue k e = case e of
                            ENum    i | isIntegralLike    -> Just $ mkConstCW k (fst i)
+                           ENum    i | isChar          k -> Just $ CW KChar    (CWChar    (chr (fromIntegral (fst i))))
                            EReal   i | isReal          k -> Just $ CW KReal    (CWAlgReal i)
                            EFloat  i | isFloat         k -> Just $ CW KFloat   (CWFloat   i)
                            EDouble i | isDouble        k -> Just $ CW KDouble  (CWDouble  i)
-                           ECon    i | isUninterpreted k -> Just $ CW k        (CWUserSort (getUIIndex k i, i))
+                           ECon    s | isString        k -> Just $ CW KString  (CWString   (interpretString s))
+                           ECon    s | isUninterpreted k -> Just $ CW k        (CWUserSort (getUIIndex k s, s))
+                           _         | isList          k -> Just $ CW k        (CWList     (interpretList e))
                            _                             -> Nothing
   where isIntegralLike = or [f k | f <- [isBoolean, isBounded, isInteger, isReal, isFloat, isDouble]]
 
         getUIIndex (KUserSort  _ (Right xs)) i = i `elemIndex` xs
         getUIIndex _                         _ = Nothing
 
+        stringLike xs = length xs >= 2 && head xs == '"' && last xs == '"'
+
+        -- Make sure strings are really strings
+        interpretString xs
+          | not (stringLike xs)
+          = error $ "Expected a string constant with quotes, received: <" ++ xs ++ ">"
+          | True
+          = qfsToString $ tail (init xs)
+
+        isStringSequence (KList (KBounded _ 8)) = True
+        isStringSequence _                      = False
+
+        -- Lists are tricky since z3 prints the 8-bit variants as strings. See: <http://github.com/Z3Prover/z3/issues/1808>
+        interpretList (ECon s)
+          | isStringSequence k && stringLike s
+          = map (CWInteger . fromIntegral . ord) $ interpretString s
+        interpretList topExpr = walk topExpr
+          where walk (EApp [ECon "as", ECon "seq.empty", _]) = []
+                walk (EApp [ECon "seq.unit", v])             = case recoverKindedValue ek v of
+                                                                 Just w -> [cwVal w]
+                                                                 Nothing -> error $ "Cannot parse a sequence item of kind " ++ show ek ++ " from: " ++ show v ++ extra v
+                walk (EApp [ECon "seq.++", pre, post])       = walk pre ++ walk post
+                walk cur                                     = error $ "Expected a sequence constant, but received: " ++ show cur ++ extra cur
+
+                extra cur | show cur == t = ""
+                          | True          = "\nWhile parsing: " ++ t
+                          where t = show topExpr
+
+                ek = case k of
+                       KList ik -> ik
+                       _        -> error $ "Impossible: Expected a sequence kind, bug got: " ++ show k
 
 -- | Get the value of a term. If the kind is Real and solver supports decimal approximations,
 -- we will "squash" the representations.
@@ -409,7 +534,10 @@ checkSat = do cfg <- getConfig
 checkSatUsing :: String -> Query CheckSatResult
 checkSatUsing cmd = do let bad = unexpected "checkSat" cmd "one of sat/unsat/unknown" Nothing
 
-                       r <- ask cmd
+                           -- Sigh.. Ignore some of the pesky warnings. We only do it as an exception here.
+                           ignoreList = ["WARNING: optimization with quantified constraints is not supported"]
+
+                       r <- askIgnoring cmd ignoreList
 
                        parse r bad $ \case ECon "sat"     -> return Sat
                                            ECon "unsat"   -> return Unsat
@@ -429,6 +557,14 @@ getQuantifiedInputs = do State{rinps} <- get
 
                          return $ preQs ++ trackers ++ postQs
 
+-- | Get observables, i.e., those explicitly labeled by the user with a call to 'Data.SBV.observe'.
+getObservables :: Query [(String, SW)]
+getObservables = do State{rObservables} <- get
+
+                    rObs <- liftIO $ readIORef rObservables
+
+                    return $ reverse rObs
+
 -- | Repeatedly issue check-sat, after refuting the previous model.
 -- The bool is true if the model is unique upto prefix existentials.
 getAllSatResult :: Query (Bool, Bool, [SMTResult])
@@ -509,7 +645,7 @@ getAllSatResult = do queryDebug ["*** Checking Satisfiability, all solutions.."]
                                                            . filter (\l -> length l > 1)  -- Only need this class if it has at least two members
                                                            . map (map fst)                -- throw away values, we only need svals
                                                            . groupBy ((==) `on` snd)      -- make sure they belong to the same sort and have the same value
-                                                           . sortBy (comparing snd)       -- sort them according to their CW (i.e., sort/value)
+                                                           . sortOn snd                   -- sort them according to their CW (i.e., sort/value)
                                                            $ uninterpreteds
                                             where pwDistinct :: [SVal] -> [SVal]
                                                   pwDistinct ss = [x `svNotEqual` y | (x:ys) <- tails ss, y <- ys]
@@ -528,8 +664,8 @@ getAllSatResult = do queryDebug ["*** Checking Satisfiability, all solutions.."]
                                         Just d  -> do constrain d
                                                       go (cnt+1) resultsSoFar
 
--- | Retrieve the set of unsatisfiable assumptions, following a call to 'checkSatAssumingWithUnsatisfiableSet'. Note that
--- this function isn't exported to the user, but rather used internally. The user simple calls 'checkSatAssumingWithUnsatisfiableSet'.
+-- | Retrieve the set of unsatisfiable assumptions, following a call to 'Data.SBV.Control.checkSatAssumingWithUnsatisfiableSet'. Note that
+-- this function isn't exported to the user, but rather used internally. The user simple calls 'Data.SBV.Control.checkSatAssumingWithUnsatisfiableSet'.
 getUnsatAssumptions :: [String] -> [(String, a)] -> Query [a]
 getUnsatAssumptions originals proxyMap = do
         let cmd = "(get-unsat-assumptions)"
@@ -602,24 +738,28 @@ unexpected ctx sent expected mbHint received mbReason = do
 
         cfg <- getConfig
 
-        let exc = SMTException { smtExceptionDescription = "Data.SBV: Unexpected response from the solver, context: " ++ ctx
-                               , smtExceptionSent        = sent
-                               , smtExceptionExpected    = expected
-                               , smtExceptionReceived    = received
-                               , smtExceptionStdOut      = unlines extras
-                               , smtExceptionStdErr      = Nothing
-                               , smtExceptionExitCode    = Nothing
-                               , smtExceptionConfig      = cfg
-                               , smtExceptionReason      = mbReason
-                               , smtExceptionHint        = mbHint
+        let exc = SBVException { sbvExceptionDescription = "Unexpected response from the solver, context: " ++ ctx
+                               , sbvExceptionSent        = Just sent
+                               , sbvExceptionExpected    = Just expected
+                               , sbvExceptionReceived    = Just received
+                               , sbvExceptionStdOut      = Just $ unlines extras
+                               , sbvExceptionStdErr      = Nothing
+                               , sbvExceptionExitCode    = Nothing
+                               , sbvExceptionConfig      = cfg
+                               , sbvExceptionReason      = mbReason
+                               , sbvExceptionHint        = mbHint
                                }
 
         io $ C.throwIO exc
 
 -- | Convert a query result to an SMT Problem
-runProofOn :: SMTConfig -> Bool -> [String] -> Result -> SMTProblem
-runProofOn config isSat comments res@(Result ki _qcInfo _codeSegs is consts tbls arrs uis axs pgm cstrs _assertions outputs) =
-     let flipQ (ALL, x) = (EX,  x)
+runProofOn :: SBVRunMode -> [String] -> Result -> SMTProblem
+runProofOn rm comments res@(Result ki _qcInfo _observables _codeSegs is consts tbls arrs uis axs pgm cstrs _assertions outputs) =
+     let (config, isSat, isSafe, isSetup) = case rm of
+                                              SMTMode stage s c -> (c, s, isSafetyCheckingIStage stage, isSetupIStage stage)
+                                              _                 -> error $ "runProofOn: Unexpected run mode: " ++ show rm
+
+         flipQ (ALL, x) = (EX,  x)
          flipQ (EX,  x) = (ALL, x)
 
          skolemize :: [(Quantifier, NamedSymVar)] -> [Either SW (SW, [SW])]
@@ -631,24 +771,145 @@ runProofOn config isSat comments res@(Result ki _qcInfo _codeSegs is consts tbls
          qinps      = if isSat then fst is else map flipQ (fst is)
          skolemMap  = skolemize qinps
 
-         o = case outputs of
-               []  -> trueSW
-               [so] -> case so of
-                        SW KBool _ -> so
-                        _          -> trueSW
-                                      {-
-                                      -- TODO: We used to error out here, but "safeWith" might have a non-bool out
-                                      -- I wish we can get rid of this and still check for it. Perhaps this entire
-                                      -- runProofOn might disappear.
-                                      error $ unlines [ "Impossible happened, non-boolean output: " ++ show so
-                                                      , "Detected while generating the trace:\n" ++ show res
-                                                      ]
-                                      -}
-               os  -> error $ unlines [ "User error: Multiple output values detected: " ++ show os
-                                      , "Detected while generating the trace:\n" ++ show res
-                                      , "*** Check calls to \"output\", they are typically not needed!"
-                                      ]
+         o | isSafe = trueSW
+           | True   = case outputs of
+                        []  | isSetup -> trueSW
+                        [so]          -> case so of
+                                           SW KBool _ -> so
+                                           _          -> error $ unlines [ "Impossible happened, non-boolean output: " ++ show so
+                                                                         , "Detected while generating the trace:\n" ++ show res
+                                                                         ]
+                        os  -> error $ unlines [ "User error: Multiple output values detected: " ++ show os
+                                               , "Detected while generating the trace:\n" ++ show res
+                                               , "*** Check calls to \"output\", they are typically not needed!"
+                                               ]
 
      in SMTProblem { smtLibPgm = toSMTLib config ki isSat comments is skolemMap consts tbls arrs uis axs pgm cstrs o }
 
+-- | Execute a query
+executeQuery :: QueryContext -> Query a -> Symbolic a
+executeQuery queryContext (Query userQuery) = do
+     st <- R.ask
+     rm <- liftIO $ readIORef (runMode st)
+
+     -- If we're doing an external query, then we cannot allow quantifiers to be present. Why?
+     -- Consider:
+     --
+     --      issue = do x :: SBool <- forall_
+     --                 y :: SBool <- exists_
+     --                 constrain y
+     --                 query $ do checkSat
+     --                         (,) <$> getValue x <*> getValue y
+     --
+     -- This is the (simplified/annotated SMTLib we would generate:)
+     --
+     --     (declare-fun s1 (Bool) Bool)   ; s1 is the function that corresponds to the skolemized 'y'
+     --     (assert (forall ((s0 Bool))    ; s0 is 'x'
+     --                 (s1 s0)))          ; s1 applied to s0 is the actual 'y'
+     --     (check-sat)
+     --     (get-value (s0))        ; s0 simply not visible here
+     --     (get-value (s1))        ; s1 is visible, but only via 's1 s0', so it is also not available.
+     --
+     -- And that would be terrible! The scoping rules of our "quantified" variables and how they map to
+     -- SMTLib is just not compatible. This is a historical design issue, but too late at this point. (We
+     -- should've never allowed general quantification like this, but only in limited contexts.)
+     --
+     -- So, we check if this is an external-query, and if there are quantified variables. If so, we
+     -- cowardly refuse to continue. For details, see: <http://github.com/LeventErkok/sbv/issues/407>
+
+     () <- liftIO $ case queryContext of
+                      QueryInternal -> return ()         -- we're good, internal usages don't mess with scopes
+                      QueryExternal -> do
+                        (userInps, _) <- readIORef (rinps st)
+                        let badInps = reverse [n | (ALL, (_, n)) <- userInps]
+                        case badInps of
+                          [] -> return ()
+                          _  -> let plu | length badInps > 1 = "s require"
+                                        | True               = " requires"
+                                in error $ unlines [ ""
+                                                   , "*** Data.SBV: Unsupported query call in the presence of quantified inputs."
+                                                   , "***"
+                                                   , "*** The following variable" ++ plu ++ " explicit quantification: "
+                                                   , "***"
+                                                   , "***    " ++ intercalate ", " badInps
+                                                   , "***"
+                                                   , "*** While quantification and queries can co-exist in principle, SBV currently"
+                                                   , "*** does not support this scenario. Avoid using quantifiers with user queries"
+                                                   , "*** if possible. Please do get in touch if your use case does require such"
+                                                   , "*** a feature to see how we can accommodate such scenarios."
+                                                   ]
+
+     case rm of
+        -- Transitioning from setup
+        SMTMode stage isSAT cfg | not (isRunIStage stage) -> liftIO $ do
+
+                                                let backend = engine (solver cfg)
+
+                                                res     <- extractSymbolicSimulationState st
+                                                setOpts <- reverse <$> readIORef (rSMTOptions st)
+
+                                                let SMTProblem{smtLibPgm} = runProofOn rm [] res
+                                                    cfg' = cfg { solverSetOptions = solverSetOptions cfg ++ setOpts }
+                                                    pgm  = smtLibPgm cfg'
+
+                                                writeIORef (runMode st) $ SMTMode IRun isSAT cfg
+
+                                                backend cfg' st (show pgm) $ evalStateT userQuery
+
+        -- Already in a query, in theory we can just continue, but that causes use-case issues
+        -- so we reject it. TODO: Review if we should actually support this. The issue arises with
+        -- expressions like this:
+        --
+        -- In the following t0's output doesn't get recorded, as the output call is too late when we get
+        -- here. (The output field isn't "incremental.") So, t0/t1 behave differently!
+        --
+        --   t0 = satWith z3{verbose=True, transcript=Just "t.smt2"} $ query (return (false::SBool))
+        --   t1 = satWith z3{verbose=True, transcript=Just "t.smt2"} $ ((return (false::SBool)) :: Predicate)
+        --
+        -- Also, not at all clear what it means to go in an out of query mode:
+        --
+        -- r = runSMTWith z3{verbose=True} $ do
+        --         a' <- sInteger "a"
+        --
+        --        (a, av) <- query $ do _ <- checkSat
+        --                              av <- getValue a'
+        --                              return (a', av)
+        --
+        --        liftIO $ putStrLn $ "Got: " ++ show av
+        --        -- constrain $ a .> literal av + 1      -- Cant' do this since we're "out" of query. Sigh.
+        --
+        --        bv <- query $ do constrain $ a .> literal av + 1
+        --                         _ <- checkSat
+        --                         getValue a
+        --
+        --        return $ a' .== a' + 1
+        --
+        -- This would be one possible implementation, alas it has the problems above:
+        --
+        --    SMTMode IRun _ _ -> liftIO $ evalStateT userQuery st
+        --
+        -- So, we just reject it.
+
+        SMTMode IRun _ _ -> error $ unlines [ ""
+                                            , "*** Data.SBV: Unsupported nested query is detected."
+                                            , "***"
+                                            , "*** Please group your queries into one block. Note that this"
+                                            , "*** can also arise if you have a call to 'query' not within 'runSMT'"
+                                            , "*** For instance, within 'sat'/'prove' calls with custom user queries."
+                                            , "*** The solution is to do the sat/prove part in the query directly."
+                                            , "***"
+                                            , "*** While multiple/nested queries should not be necessary in general,"
+                                            , "*** please do get in touch if your use case does require such a feature,"
+                                            , "*** to see how we can accommodate such scenarios."
+                                            ]
+
+        -- Otherwise choke!
+        m -> error $ unlines [ ""
+                             , "*** Data.SBV: Invalid query call."
+                             , "***"
+                             , "***   Current mode: " ++ show m
+                             , "***"
+                             , "*** Query calls are only valid within runSMT/runSMTWith calls"
+                             ]
+
 {-# ANN module ("HLint: ignore Reduce duplication" :: String) #-}
