@@ -18,9 +18,7 @@ module Data.SBV.Core.Floating (
        , blastSFloat, blastSDouble
        ) where
 
-import Control.Monad (join)
-
-import qualified Data.Binary.IEEE754 as DB (wordToFloat, wordToDouble, floatToWord, doubleToWord)
+import qualified Data.ReinterpretCast as RC (wordToFloat, wordToDouble, floatToWord, doubleToWord)
 
 import Data.Int            (Int8,  Int16,  Int32,  Int64)
 import Data.Word           (Word8, Word16, Word32, Word64)
@@ -249,7 +247,7 @@ instance IEEEFloatConvertable AlgReal where
 concEval1 :: SymWord a => Maybe (a -> a) -> Maybe SRoundingMode -> SBV a -> Maybe (SBV a)
 concEval1 mbOp mbRm a = do op <- mbOp
                            v  <- unliteral a
-                           case join (unliteral `fmap` mbRm) of
+                           case unliteral =<< mbRm of
                              Nothing                     -> (Just . literal) (op v)
                              Just RoundNearestTiesToEven -> (Just . literal) (op v)
                              _                           -> Nothing
@@ -259,7 +257,7 @@ concEval2 :: SymWord a => Maybe (a -> a -> a) -> Maybe SRoundingMode -> SBV a ->
 concEval2 mbOp mbRm a b  = do op <- mbOp
                               v1 <- unliteral a
                               v2 <- unliteral b
-                              case join (unliteral `fmap` mbRm) of
+                              case unliteral =<< mbRm of
                                 Nothing                     -> (Just . literal) (v1 `op` v2)
                                 Just RoundNearestTiesToEven -> (Just . literal) (v1 `op` v2)
                                 _                           -> Nothing
@@ -269,7 +267,7 @@ concEval2B :: SymWord a => Maybe (a -> a -> Bool) -> Maybe SRoundingMode -> SBV
 concEval2B mbOp mbRm a b  = do op <- mbOp
                                v1 <- unliteral a
                                v2 <- unliteral b
-                               case join (unliteral `fmap` mbRm) of
+                               case unliteral =<< mbRm of
                                  Nothing                     -> (Just . literal) (v1 `op` v2)
                                  Just RoundNearestTiesToEven -> (Just . literal) (v1 `op` v2)
                                  _                           -> Nothing
@@ -280,7 +278,7 @@ concEval3 mbOp mbRm a b c = do op <- mbOp
                                v1 <- unliteral a
                                v2 <- unliteral b
                                v3 <- unliteral c
-                               case join (unliteral `fmap` mbRm) of
+                               case unliteral =<< mbRm of
                                  Nothing                     -> (Just . literal) (op v1 v2 v3)
                                  Just RoundNearestTiesToEven -> (Just . literal) (op v1 v2 v3)
                                  _                           -> Nothing
@@ -381,7 +379,7 @@ lift3 w mbOp mbRm a b c
 sFloatAsSWord32 :: SFloat -> SWord32
 sFloatAsSWord32 fVal
   | Just f <- unliteral fVal, not (isNaN f)
-  = literal (DB.floatToWord f)
+  = literal (RC.floatToWord f)
   | True
   = SBV (SVal w32 (Right (cache y)))
   where w32  = KBounded False 32
@@ -391,7 +389,7 @@ sFloatAsSWord32 fVal
                              newExpr st w32 (SBVApp (IEEEFP (FP_Reinterpret KFloat w32)) [f])
                      else do n   <- internalVariable st w32
                              ysw <- newExpr st KFloat (SBVApp (IEEEFP (FP_Reinterpret w32 KFloat)) [n])
-                             internalConstraint st Nothing $ unSBV $ fVal `fpIsEqualObject` SBV (SVal KFloat (Right (cache (\_ -> return ysw))))
+                             internalConstraint st False [] $ unSBV $ fVal `fpIsEqualObject` SBV (SVal KFloat (Right (cache (\_ -> return ysw))))
                              return n
 
 -- | Convert an 'SDouble' to an 'SWord64', preserving the bit-correspondence. Note that since the
@@ -402,7 +400,7 @@ sFloatAsSWord32 fVal
 sDoubleAsSWord64 :: SDouble -> SWord64
 sDoubleAsSWord64 fVal
   | Just f <- unliteral fVal, not (isNaN f)
-  = literal (DB.doubleToWord f)
+  = literal (RC.doubleToWord f)
   | True
   = SBV (SVal w64 (Right (cache y)))
   where w64  = KBounded False 64
@@ -412,7 +410,7 @@ sDoubleAsSWord64 fVal
                              newExpr st w64 (SBVApp (IEEEFP (FP_Reinterpret KDouble w64)) [f])
                      else do n   <- internalVariable st w64
                              ysw <- newExpr st KDouble (SBVApp (IEEEFP (FP_Reinterpret w64 KDouble)) [n])
-                             internalConstraint st Nothing $ unSBV $ fVal `fpIsEqualObject` SBV (SVal KDouble (Right (cache (\_ -> return ysw))))
+                             internalConstraint st False [] $ unSBV $ fVal `fpIsEqualObject` SBV (SVal KDouble (Right (cache (\_ -> return ysw))))
                              return n
 
 -- | Extract the sign\/exponent\/mantissa of a single-precision float. The output will have
@@ -430,7 +428,7 @@ blastSDouble = extract . sDoubleAsSWord64
 -- | Reinterpret the bits in a 32-bit word as a single-precision floating point number
 sWord32AsSFloat :: SWord32 -> SFloat
 sWord32AsSFloat fVal
-  | Just f <- unliteral fVal = literal $ DB.wordToFloat f
+  | Just f <- unliteral fVal = literal $ RC.wordToFloat f
   | True                     = SBV (SVal KFloat (Right (cache y)))
   where y st = do xsw <- sbvToSW st fVal
                   newExpr st KFloat (SBVApp (IEEEFP (FP_Reinterpret (kindOf fVal) KFloat)) [xsw])
@@ -438,7 +436,7 @@ sWord32AsSFloat fVal
 -- | Reinterpret the bits in a 32-bit word as a single-precision floating point number
 sWord64AsSDouble :: SWord64 -> SDouble
 sWord64AsSDouble dVal
-  | Just d <- unliteral dVal = literal $ DB.wordToDouble d
+  | Just d <- unliteral dVal = literal $ RC.wordToDouble d
   | True                     = SBV (SVal KDouble (Right (cache y)))
   where y st = do xsw <- sbvToSW st dVal
                   newExpr st KDouble (SBVApp (IEEEFP (FP_Reinterpret (kindOf dVal) KDouble)) [xsw])
