@@ -9,32 +9,29 @@
 -- Instance declarations for our symbolic world
 -----------------------------------------------------------------------------
 
-{-# LANGUAGE TypeSynonymInstances   #-}
-{-# LANGUAGE BangPatterns           #-}
-{-# LANGUAGE PatternGuards          #-}
-{-# LANGUAGE FlexibleContexts       #-}
-{-# LANGUAGE FlexibleInstances      #-}
-{-# LANGUAGE MultiParamTypeClasses  #-}
-{-# LANGUAGE ScopedTypeVariables    #-}
-{-# LANGUAGE Rank2Types             #-}
-{-# LANGUAGE TypeOperators          #-}
-{-# LANGUAGE DefaultSignatures      #-}
+{-# LANGUAGE BangPatterns          #-}
+{-# LANGUAGE DefaultSignatures     #-}
+{-# LANGUAGE FlexibleContexts      #-}
+{-# LANGUAGE FlexibleInstances     #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE PatternGuards         #-}
+{-# LANGUAGE ScopedTypeVariables   #-}
+{-# LANGUAGE Rank2Types            #-}
+{-# LANGUAGE TypeOperators         #-}
+{-# LANGUAGE TypeSynonymInstances  #-}
 
--- Keep GHC quiet.. I wish we can be more specific on these as opposed to module
--- level turning of redundant-constraints. Note that his is not needed >= GHC 8.0.2
--- since they removed this constraint from -Wall, see: https://ghc.haskell.org/trac/ghc/ticket/10635
-{-# OPTIONS_GHC -fno-warn-orphans               #-}
-{-# OPTIONS_GHC -fno-warn-redundant-constraints #-}
+{-# OPTIONS_GHC -fno-warn-orphans  #-}
 
 module Data.SBV.Core.Model (
-    Mergeable(..), EqSymbolic(..), OrdSymbolic(..), SDivisible(..), Uninterpreted(..), Metric(..), assertSoft, SIntegral, SFiniteBits(..)
+    Mergeable(..), EqSymbolic(..), OrdSymbolic(..), SDivisible(..), Uninterpreted(..), Metric(..), assertWithPenalty, SIntegral, SFiniteBits(..)
   , ite, iteLazy, sFromIntegral, sShiftLeft, sShiftRight, sRotateLeft, sRotateRight, sSignedShiftArithRight, (.^)
   , oneIf, genVar, genVar_, forall, forall_, exists, exists_
   , pbAtMost, pbAtLeast, pbExactly, pbLe, pbGe, pbEq, pbMutexed, pbStronglyMutexed
   , sBool, sBools, sWord8, sWord8s, sWord16, sWord16s, sWord32
   , sWord32s, sWord64, sWord64s, sInt8, sInt8s, sInt16, sInt16s, sInt32, sInt32s, sInt64
-  , sInt64s, sInteger, sIntegers, sReal, sReals, sFloat, sFloats, sDouble, sDoubles, slet
-  , sRealToSInteger, label
+  , sInt64s, sInteger, sIntegers, sReal, sReals, sFloat, sFloats, sDouble, sDoubles, sChar, sChars, sString, sStrings, sList, sLists
+  , slet
+  , sRealToSInteger, label, observe
   , sAssert
   , liftQRem, liftDMod, symbolicMergeWithKind
   , genLiteral, genFromCW, genMkSymVar
@@ -42,19 +39,26 @@ module Data.SBV.Core.Model (
   )
   where
 
+import Control.Applicative  (ZipList(ZipList))
 import Control.Monad        (when, unless, mplus)
+import Control.Monad.Trans  (liftIO)
+import Control.Monad.Reader (ask)
 
 import GHC.Generics (U1(..), M1(..), (:*:)(..), K1(..))
 import qualified GHC.Generics as G
 
 import GHC.Stack
 
-import Data.Array      (Array, Ix, listArray, elems, bounds, rangeSize)
-import Data.Bits       (Bits(..))
-import Data.Int        (Int8, Int16, Int32, Int64)
-import Data.List       (genericLength, genericIndex, genericTake, unzip4, unzip5, unzip6, unzip7, intercalate)
-import Data.Maybe      (fromMaybe)
-import Data.Word       (Word8, Word16, Word32, Word64)
+import Data.Array  (Array, Ix, listArray, elems, bounds, rangeSize)
+import Data.Bits   (Bits(..))
+import Data.Char   (toLower, isDigit)
+import Data.Int    (Int8, Int16, Int32, Int64)
+import Data.List   (genericLength, genericIndex, genericTake, unzip4, unzip5, unzip6, unzip7, intercalate, isPrefixOf)
+import Data.Maybe  (fromMaybe)
+import Data.String (IsString(..))
+import Data.Word   (Word8, Word16, Word32, Word64)
+
+import Data.Dynamic (fromDynamic, toDyn)
 
 import Test.QuickCheck                         (Testable(..), Arbitrary(..))
 import qualified Test.QuickCheck.Test    as QC (isSuccess)
@@ -66,16 +70,11 @@ import Data.SBV.Core.Data
 import Data.SBV.Core.Symbolic
 import Data.SBV.Core.Operations
 
-import Data.SBV.Provers.Prover (defaultSMTCfg)
+import Data.SBV.Provers.Prover (defaultSMTCfg, SafeResult(..))
 import Data.SBV.SMT.SMT        (showModel)
 
 import Data.SBV.Utils.Boolean
-
-mkSymOpSC :: (SW -> SW -> Maybe SW) -> Op -> State -> Kind -> SW -> SW -> IO SW
-mkSymOpSC shortCut op st k a b = maybe (newExpr st k (SBVApp op [a, b])) return (shortCut a b)
-
-mkSymOp :: Op -> State -> Kind -> SW -> SW -> IO SW
-mkSymOp = mkSymOpSC (const (const Nothing))
+import Data.SBV.Utils.Lib      (isKString)
 
 -- Symbolic-Word class instances
 
@@ -186,9 +185,38 @@ instance SymWord Double where
   -- and in the presence of NaN's it would be incorrect to do any optimization
   isConcretely _ _ = False
 
+instance SymWord Char where
+  mkSymWord = genMkSymVar KChar
+  literal c = SBV . SVal KChar . Left . CW KChar $ CWChar c
+  fromCW (CW _ (CWChar a)) = a
+  fromCW c                 = error $ "SymWord.String: Unexpected non-char value: " ++ show c
+
+instance SymWord a => SymWord [a] where
+  mkSymWord
+    | isKString (undefined :: [a]) = genMkSymVar KString
+    | True                         = genMkSymVar (KList (kindOf (undefined :: a)))
+
+  literal as
+    | isKString (undefined :: [a]) = case fromDynamic (toDyn as) of
+                                       Just s  -> SBV . SVal KString . Left . CW KString . CWString $ s
+                                       Nothing -> error "SString: Cannot construct literal string!"
+    | True                         = let k = KList (kindOf (undefined :: a))
+                                         toCWVal a = case literal a of
+                                                       SBV (SVal _ (Left (CW _ cwval))) -> cwval
+                                                       _                                -> error "SymWord.Sequence: could not produce a concrete word for value"
+                                     in SBV $ SVal k $ Left $ CW k $ CWList $ map toCWVal as
+
+  fromCW (CW _ (CWString a)) = fromMaybe (error "SString: Cannot extract a literal string!")
+                                         (fromDynamic (toDyn a))
+  fromCW (CW _ (CWList a))   = fromCW . CW (kindOf (undefined :: a)) <$> a
+  fromCW c                   = error $ "SymWord.fromCW: Unexpected non-list value: " ++ show c
+
+instance IsString SString where
+  fromString = literal
+
 ------------------------------------------------------------------------------------
 -- * Smart constructors for creating symbolic values. These are not strictly
--- necessary, as they are mere aliases for 'symbolic' and 'symbolics', but 
+-- necessary, as they are mere aliases for 'symbolic' and 'symbolics', but
 -- they nonetheless make programming easier.
 ------------------------------------------------------------------------------------
 -- | Declare an 'SBool'
@@ -295,6 +323,30 @@ sDouble = symbolic
 sDoubles :: [String] -> Symbolic [SDouble]
 sDoubles = symbolics
 
+-- | Declare an 'SChar'
+sChar :: String -> Symbolic SChar
+sChar = symbolic
+
+-- | Declare an 'SString'
+sString :: String -> Symbolic SString
+sString = symbolic
+
+-- | Declare a list of 'SChar's
+sChars :: [String] -> Symbolic [SChar]
+sChars = symbolics
+
+-- | Declare a list of 'SString's
+sStrings :: [String] -> Symbolic [SString]
+sStrings = symbolics
+
+-- | Declare an 'SList'
+sList :: forall a. SymWord a => String -> Symbolic (SList a)
+sList = symbolic
+
+-- | Declare a list of 'SList's
+sLists :: forall a. SymWord a => [String] -> Symbolic [SList a]
+sLists = symbolics
+
 -- | Convert an SReal to an SInteger. That is, it computes the
 -- largest integer @n@ that satisfies @sIntegerToSReal n <= r@
 -- essentially giving us the @floor@.
@@ -319,6 +371,22 @@ label m x
         r st = do xsw <- sbvToSW st x
                   newExpr st k (SBVApp (Label m) [xsw])
 
+-- | Observe the value of an expression.  Such values are useful in model construction, as they are printed part of a satisfying model, or a
+-- counter-example. The same works for quick-check as well. Useful when we want to see intermediate values, or expected/obtained
+-- pairs in a particular run.
+observe :: SymWord a => String -> SBV a -> Symbolic ()
+observe m x
+  | null m
+  = error "SBV.observe: Bad empty name!"
+  | map toLower m `elem` smtLibReservedNames
+  = error $ "SBV.observe: The name chosen is reserved, please change it!: " ++ show m
+  | "s" `isPrefixOf` m && all isDigit (drop 1 m)
+  = error $ "SBV.observe: Names of the form sXXX are internal to SBV, please use a different name: " ++ show m
+  | True
+  = do st <- ask
+       liftIO $ do xsw <- sbvToSW st x
+                   recordObservable st m xsw
+
 -- | Symbolic Equality. Note that we can't use Haskell's 'Eq' class since Haskell insists on returning Bool
 -- Comparing symbolic values will necessarily return a symbolic value.
 infix 4 .==, ./=
@@ -336,8 +404,8 @@ class EqSymbolic a where
 
   -- | Symbolic membership test.
   sElem    :: a -> [a] -> SBool
+  {-# MINIMAL (.==) #-}
 
-  -- minimal complete definition: .==
   x ./= y = bnot (x .== y)
 
   allEqual []     = true
@@ -371,7 +439,8 @@ class (Mergeable a, EqSymbolic a) => OrdSymbolic a where
   -- | Is the value withing the allowed /inclusive/ range?
   inRange    :: a -> (a, a) -> SBool
 
-  -- minimal complete definition: .<
+  {-# MINIMAL (.<) #-}
+
   a .<= b    = a .< b ||| a .== b
   a .>  b    = b .<  a
   a .>= b    = b .<= a
@@ -536,7 +605,7 @@ instance SIntegral Int64
 instance SIntegral Integer
 
 -- | Finite bit-length symbolic values. Essentially the same as 'SIntegral', but further leaves out 'Integer'. Loosely
--- based on Haskell's 'FiniteBits' class, but with more methods defined and structured differently to fit into the
+-- based on Haskell's @FiniteBits@ class, but with more methods defined and structured differently to fit into the
 -- symbolic world view. Minimal complete definition: 'sFiniteBitSize'.
 class (SymWord a, Num a, Bits a) => SFiniteBits a where
     -- | Bit size.
@@ -691,42 +760,42 @@ liftPB w o xs
                     registerKind st KUnbounded
                     newExpr st KBool (SBVApp (PseudoBoolean o) xsw)
 
--- | 'true' if at most `k` of the input arguments are 'true'
+-- | 'true' if at most @k@ of the input arguments are 'true'
 pbAtMost :: [SBool] -> Int -> SBool
 pbAtMost xs k
  | k < 0             = error $ "SBV.pbAtMost: Non-negative value required, received: " ++ show k
  | all isConcrete xs = literal $ sum (map (pbToInteger "pbAtMost" 1) xs) <= fromIntegral k
  | True              = liftPB "pbAtMost" (PB_AtMost k) xs
 
--- | 'true' if at least `k` of the input arguments are 'true'
+-- | 'true' if at least @k@ of the input arguments are 'true'
 pbAtLeast :: [SBool] -> Int -> SBool
 pbAtLeast xs k
  | k < 0             = error $ "SBV.pbAtLeast: Non-negative value required, received: " ++ show k
  | all isConcrete xs = literal $ sum (map (pbToInteger "pbAtLeast" 1) xs) >= fromIntegral k
  | True              = liftPB "pbAtLeast" (PB_AtLeast k) xs
 
--- | 'true' if exactly `k` of the input arguments are 'true'
+-- | 'true' if exactly @k@ of the input arguments are 'true'
 pbExactly :: [SBool] -> Int -> SBool
 pbExactly xs k
  | k < 0             = error $ "SBV.pbExactly: Non-negative value required, received: " ++ show k
  | all isConcrete xs = literal $ sum (map (pbToInteger "pbExactly" 1) xs) == fromIntegral k
  | True              = liftPB "pbExactly" (PB_Exactly k) xs
 
--- | 'true' if the sum of coefficients for 'true' elements is at most 'k'. Generalizes 'pbAtMost'.
+-- | 'true' if the sum of coefficients for 'true' elements is at most @k@. Generalizes 'pbAtMost'.
 pbLe :: [(Int, SBool)] -> Int -> SBool
 pbLe xs k
  | k < 0                       = error $ "SBV.pbLe: Non-negative value required, received: " ++ show k
  | all isConcrete (map snd xs) = literal $ sum [pbToInteger "pbLe" c b | (c, b) <- xs] <= fromIntegral k
  | True                        = liftPB "pbLe" (PB_Le (map fst xs) k) (map snd xs)
 
--- | 'true' if the sum of coefficients for 'true' elements is at least 'k'. Generalizes 'pbAtLeast'.
+-- | 'true' if the sum of coefficients for 'true' elements is at least @k@. Generalizes 'pbAtLeast'.
 pbGe :: [(Int, SBool)] -> Int -> SBool
 pbGe xs k
  | k < 0                       = error $ "SBV.pbGe: Non-negative value required, received: " ++ show k
  | all isConcrete (map snd xs) = literal $ sum [pbToInteger "pbGe" c b | (c, b) <- xs] >= fromIntegral k
  | True                        = liftPB "pbGe" (PB_Ge (map fst xs) k) (map snd xs)
 
--- | 'true' if the sum of coefficients for 'true' elements is exactly least 'k'. Useful for coding
+-- | 'true' if the sum of coefficients for 'true' elements is exactly least @k@. Useful for coding
 -- /exactly K-of-N/ constraints, and in particular mutex constraints.
 pbEq :: [(Int, SBool)] -> Int -> SBool
 pbEq xs k
@@ -820,6 +889,9 @@ instance (SymWord a, Fractional a) => Fractional (SBV a) where
                       k@KBounded{}  -> error $ "Unexpected Fractional case for: " ++ show k
                       k@KUnbounded  -> error $ "Unexpected Fractional case for: " ++ show k
                       k@KBool       -> error $ "Unexpected Fractional case for: " ++ show k
+                      k@KString     -> error $ "Unexpected Fractional case for: " ++ show k
+                      k@KChar       -> error $ "Unexpected Fractional case for: " ++ show k
+                      k@KList{}     -> error $ "Unexpected Fractional case for: " ++ show k
                       k@KUserSort{} -> error $ "Unexpected Fractional case for: " ++ show k
 
 -- | Define Floating instance on SBV's; only for base types that are already floating; i.e., SFloat and SDouble
@@ -901,7 +973,7 @@ instance (Num a, Bits a, SymWord a) => Bits (SBV a) where
     | True
     = error $ "SBV.popCount: Called on symbolic value: " ++ show x ++ ". Use sPopCount instead."
 
--- | Conversion between integral-symbolic values, akin to Haskell's fromIntegral
+-- | Conversion between integral-symbolic values, akin to Haskell's `fromIntegral`
 sFromIntegral :: forall a b. (Integral a, HasKind a, Num a, SymWord a, HasKind b, Num b, SymWord b) => SBV a -> SBV b
 sFromIntegral x
   | isReal x
@@ -1019,7 +1091,7 @@ enumCvt w x = case unliteral x of
 -- | The 'SDivisible' class captures the essence of division.
 -- Unfortunately we cannot use Haskell's 'Integral' class since the 'Real'
 -- and 'Enum' superclasses are not implementable for symbolic bit-vectors.
--- However, 'quotRem' and 'divMod' makes perfect sense, and the 'SDivisible' class captures
+-- However, 'quotRem' and 'divMod' both make perfect sense, and the 'SDivisible' class captures
 -- this operation. One issue is how division by 0 behaves. The verification
 -- technology requires total functions, and there are several design choices
 -- here. We follow Isabelle/HOL approach of assigning the value 0 for division
@@ -1144,7 +1216,7 @@ instance SDivisible SInt8 where
   sQuotRem = liftQRem
   sDivMod  = liftDMod
 
--- | Lift 'QRem' to symbolic words. Division by 0 is defined s.t. @x/0 = 0@; which
+-- | Lift 'quotRem' to symbolic words. Division by 0 is defined s.t. @x/0 = 0@; which
 -- holds even when @x@ is @0@ itself.
 liftQRem :: SymWord a => SBV a -> SBV a -> (SBV a, SBV a)
 liftQRem x y
@@ -1167,7 +1239,7 @@ liftQRem x y
                                    mkSymOp o st sgnsz sw1 sw2
         z = genLiteral (kindOf x) (0::Integer)
 
--- | Lift 'DMod' to symbolic words. Division by 0 is defined s.t. @x/0 = 0@; which
+-- | Lift 'divMod' to symbolic words. Division by 0 is defined s.t. @x/0 = 0@; which
 -- holds even when @x@ is @0@ itself. Essentially, this is conversion from quotRem
 -- (truncate to 0) to divMod (truncate towards negative infinity)
 liftDMod :: (SymWord a, Num a, SDivisible (SBV a)) => SBV a -> SBV a -> (SBV a, SBV a)
@@ -1204,12 +1276,6 @@ instance SDivisible SInteger where
             $ ite (y .>  0)              1 (-1)
 
 -- Quickcheck interface
-
--- The Arbitrary instance for SFunArray returns an array initialized
--- to an arbitrary element
-instance (SymWord b, Arbitrary b) => Arbitrary (SFunArray a b) where
-  arbitrary = arbitrary >>= \r -> return $ SFunArray (const r)
-
 instance (SymWord a, Arbitrary a) => Arbitrary (SBV a) where
   arbitrary = literal `fmap` arbitrary
 
@@ -1222,16 +1288,16 @@ instance (SymWord a, Arbitrary a) => Arbitrary (SBV a) where
 -- with a single constructor where the type of each field is an instance of
 -- 'Mergeable', such as a record of symbolic values. Users only need to add
 -- 'G.Generic' and 'Mergeable' to the @deriving@ clause for the data-type. See
--- 'Data.SBV.Examples.Puzzles.U2Bridge.Status' for an example and an
+-- 'Documentation.SBV.Examples.Puzzles.U2Bridge.Status' for an example and an
 -- illustration of what the instance would look like if written by hand.
 --
 -- The function 'select' is a total-indexing function out of a list of choices
 -- with a default value, simulating array/list indexing. It's an n-way generalization
 -- of the 'ite' function.
 --
--- Minimal complete definition: None, if the type is instance of 'Generic'. Otherwise
+-- Minimal complete definition: None, if the type is instance of @Generic@. Otherwise
 -- 'symbolicMerge'. Note that most types subject to merging are likely to be
--- trivial instances of 'Generic'.
+-- trivial instances of @Generic@.
 class Mergeable a where
    -- | Merge two values based on the condition. The first argument states
    -- whether we force the then-and-else branches before the merging, at the
@@ -1264,7 +1330,6 @@ class Mergeable a where
    default symbolicMerge :: (G.Generic a, GMergeable (G.Rep a)) => Bool -> SBool -> a -> a -> a
    symbolicMerge = symbolicMergeDefault
 
-
 -- | If-then-else. This is by definition 'symbolicMerge' with both
 -- branches forced. This is typically the desired behavior, but also
 -- see 'iteLazy' should you need more laziness.
@@ -1284,7 +1349,13 @@ iteLazy t a b
 -- | Symbolic assert. Check that the given boolean condition is always true in the given path. The
 -- optional first argument can be used to provide call-stack info via GHC's location facilities.
 sAssert :: Maybe CallStack -> String -> SBool -> SBV a -> SBV a
-sAssert cs msg cond x = SBV $ SVal k $ Right $ cache r
+sAssert cs msg cond x
+   | Just mustHold <- unliteral cond
+   = if mustHold
+     then x
+     else error $ show $ SafeResult ((locInfo . getCallStack) `fmap` cs, msg, Satisfiable defaultSMTCfg (SMTModel [] []))
+   | True
+   = SBV $ SVal k $ Right $ cache r
   where k     = kindOf x
         r st  = do xsw <- sbvToSW st x
                    let pc = getPathCondition st
@@ -1295,6 +1366,9 @@ sAssert cs msg cond x = SBV $ SVal k $ Right $ cache r
                    addAssertion st cs msg cnd
                    return xsw
 
+        locInfo ps = intercalate ",\n " (map loc ps)
+          where loc (f, sl) = concat [srcLocFile sl, ":", show (srcLocStartLine sl), ":", show (srcLocStartCol sl), ":", f]
+
 -- | Merge two symbolic values, at kind @k@, possibly @force@'ing the branches to make
 -- sure they do not evaluate to the same result. This should only be used for internal purposes;
 -- as default definitions provided should suffice in many cases. (i.e., End users should
@@ -1350,6 +1424,11 @@ instance Mergeable a => Mergeable [a] where
     | True       = error $ "SBV.Mergeable.List: No least-upper-bound for lists of differing size " ++ show (lxs, lys)
     where (lxs, lys) = (length xs, length ys)
 
+-- ZipList
+instance Mergeable a => Mergeable (ZipList a) where
+  symbolicMerge force test (ZipList xs) (ZipList ys)
+    = ZipList (symbolicMerge force test xs ys)
+
 -- Maybe
 instance Mergeable a => Mergeable (Maybe a) where
   symbolicMerge _ _ Nothing  Nothing  = Nothing
@@ -1467,23 +1546,13 @@ instance (SymWord a, Bounded a) => Bounded (SBV a) where
 
 -- SArrays are both "EqSymbolic" and "Mergeable"
 instance EqSymbolic (SArray a b) where
-  (SArray a) .== (SArray b) = SBV (eqSArr a b)
+  SArray a .== SArray b = SBV (a `eqSArr` b)
 
 -- When merging arrays; we'll ignore the force argument. This is arguably
 -- the right thing to do as we've too many things and likely we want to keep it efficient.
 instance SymWord b => Mergeable (SArray a b) where
   symbolicMerge _ = mergeArrays
 
--- SFunArrays are only "Mergeable". Although a brute
--- force equality can be defined, any non-toy instance
--- will suffer from efficiency issues; so we don't define it
-instance SymArray SFunArray where
-  newArray nm                                 = declNewSFunArray (Just nm)
-  newArray_                                   = declNewSFunArray Nothing
-  readArray     (SFunArray f)                 = f
-  writeArray    (SFunArray f) a b             = SFunArray (\a' -> ite (a .== a') b (f a'))
-  mergeArrays t (SFunArray g)   (SFunArray h) = SFunArray (\x -> ite t (g x) (h x))
-
 -- When merging arrays; we'll ignore the force argument. This is arguably
 -- the right thing to do as we've too many things and likely we want to keep it efficient.
 instance SymWord b => Mergeable (SFunArray a b) where
@@ -1513,7 +1582,9 @@ class Uninterpreted a where
   -- by end-user-code, but is rather useful for the library development.
   sbvUninterpret :: Maybe ([String], a) -> String -> a
 
-  -- minimal complete definition: 'sbvUninterpret'
+  {-# MINIMAL sbvUninterpret #-}
+
+  -- defaults:
   uninterpret             = sbvUninterpret Nothing
   cgUninterpret nm code v = sbvUninterpret (Just (code, v)) nm
 
@@ -1734,13 +1805,16 @@ instance (SymWord h, SymWord g, SymWord f, SymWord e, SymWord d, SymWord c, SymW
 
 -- | Symbolic computations provide a context for writing symbolic programs.
 instance SolverContext Symbolic where
-   constrain          (SBV c) = imposeConstraint Nothing   c
-   namedConstraint nm (SBV c) = imposeConstraint (Just nm) c
-   setOption o                = addNewSMTOption  o
+   constrain                   (SBV c) = imposeConstraint False []               c
+   softConstrain               (SBV c) = imposeConstraint True  []               c
+   namedConstraint        nm   (SBV c) = imposeConstraint False [(":named", nm)] c
+   constrainWithAttribute atts (SBV c) = imposeConstraint False atts             c
+
+   setOption o = addNewSMTOption  o
 
 -- | Introduce a soft assertion, with an optional penalty
-assertSoft :: String -> SBool -> Penalty -> Symbolic ()
-assertSoft nm o p = addSValOptGoal $ unSBV `fmap` AssertSoft nm o p
+assertWithPenalty :: String -> SBool -> Penalty -> Symbolic ()
+assertWithPenalty nm o p = addSValOptGoal $ unSBV `fmap` AssertWithPenalty nm o p
 
 -- | Class of metrics we can optimize for. Currently,
 -- bounded signed/unsigned bit-vectors, unbounded integers,
@@ -1748,7 +1822,7 @@ assertSoft nm o p = addSValOptGoal $ unSBV `fmap` AssertSoft nm o p
 -- Minimal complete definition: minimize/maximize.
 --
 -- A good reference on these features is given in the following paper:
--- <http://www.easychair.org/publications/download/Z_-_Maximal_Satisfaction_with_Z3>.
+-- <http://www.microsoft.com/en-us/research/wp-content/uploads/2016/02/nbjorner-scss2014.pdf>.
 class Metric a where
   -- | Minimize a named metric
   minimize :: String -> a -> Symbolic ()
@@ -1770,29 +1844,33 @@ instance Metric SReal    where minimize nm o = addSValOptGoal (unSBV `fmap` Mini
 -- Quickcheck interface on symbolic-booleans..
 instance Testable SBool where
   property (SBV (SVal _ (Left b))) = property (cwToBool b)
-  property s                       = error $ "Cannot quick-check in the presence of uninterpreted constants! (" ++ show s ++ ")"
+  property _                       = error "Quick-check: Constant folding produced a symbolic value! Perhaps used a non-reducible expression? Please report!"
 
 instance Testable (Symbolic SBool) where
-   property prop = QC.monadicIO $ do (cond, r, tvals) <- QC.run test
+   property prop = QC.monadicIO $ do (cond, r, modelVals) <- QC.run test
                                      QC.pre cond
-                                     unless (r || null tvals) $ QC.monitor (QC.counterexample (complain tvals))
+                                     unless (r || null modelVals) $ QC.monitor (QC.counterexample (complain modelVals))
                                      QC.assert r
-     where test = do (r, Result{resTraces=tvals, resConsts=cs, resConstraints=cstrs, resUIConsts=unints}) <- runSymbolic Concrete prop
+     where test = do (r, Result{resTraces=tvals, resObservables=ovals, resConsts=cs, resConstraints=cstrs, resUIConsts=unints}) <- runSymbolic Concrete prop
 
                      let cval = fromMaybe (error "Cannot quick-check in the presence of uninterpeted constants!") . (`lookup` cs)
-                         cond = all (cwToBool . cval . snd) cstrs
+                         cond = and [cwToBool (cval v) | (False, _, v) <- cstrs] -- Only pick-up "hard" constraints, as indicated by False in the fist component
+
+                         getObservable (nm, v) = case v `lookup` cs of
+                                                   Just cw -> (nm, cw)
+                                                   Nothing -> error $ "Quick-check: Observable " ++ nm ++ " did not reduce to a constant!"
 
                      case map fst unints of
                        [] -> case unliteral r of
                                Nothing -> noQC [show r]
-                               Just b  -> return (cond, b, tvals)
+                               Just b  -> return (cond, b, tvals ++ map getObservable ovals)
                        us -> noQC us
 
            complain qcInfo = showModel defaultSMTCfg (SMTModel [] qcInfo)
 
            noQC us         = error $ "Cannot quick-check in the presence of uninterpreted constants: " ++ intercalate ", " us
 
--- | Quick check an SBV property. Note that a regular 'quickCheck' call will work just as
+-- | Quick check an SBV property. Note that a regular @quickCheck@ call will work just as
 -- well. Use this variant if you want to receive the boolean result.
 sbvQuickCheck :: Symbolic SBool -> IO Bool
 sbvQuickCheck prop = QC.isSuccess `fmap` QC.quickCheckResult prop
@@ -1805,7 +1883,7 @@ instance Testable (Symbolic SVal) where
                              return (SBV s :: SBool)
 
 -- | Explicit sharing combinator. The SBV library has internal caching/hash-consing mechanisms
--- built in, based on Andy Gill's type-safe obervable sharing technique (see: <http://ittc.ku.edu/~andygill/paper.php?label=DSLExtract09>).
+-- built in, based on Andy Gill's type-safe obervable sharing technique (see: <http://ku-fpg.github.io/files/Gill-09-TypeSafeReification.pdf>).
 -- However, there might be times where being explicit on the sharing can help, especially in experimental code. The 'slet' combinator
 -- ensures that its first argument is computed once and passed on to its continuation, explicitly indicating the intent of sharing. Most
 -- use cases of the SBV library should simply use Haskell's @let@ construct for this purpose.
