@@ -22,7 +22,7 @@ module Data.SBV.Internals (
   , module Data.SBV.Core.Data
 
   -- * Operations useful for instantiating SBV type classes
-  , genLiteral, genFromCW, CW(..), genMkSymVar, genParse, showModel, SMTModel(..), liftQRem, liftDMod
+  , genLiteral, genFromCW, CW(..), genMkSymVar, genParse, showModel, SMTModel(..), liftQRem, liftDMod, registerKind
 
   -- * Compilation to C, extras
   , compileToC', compileToCLib'
@@ -43,11 +43,14 @@ module Data.SBV.Internals (
   -- $coordinateSolverInfo
   , sendStringToSolver, sendRequestToSolver, retrieveResponseFromSolver
 
+  -- * Defining new metrics
+  , addSValOptGoal
+
   ) where
 
 import Data.SBV.Core.Data
 import Data.SBV.Core.Model      (genLiteral, genFromCW, genMkSymVar, liftQRem, liftDMod)
-import Data.SBV.Core.Symbolic   (IStage(..))
+import Data.SBV.Core.Symbolic   (IStage(..), addSValOptGoal, registerKind)
 
 import Data.SBV.Compilers.C       (compileToC', compileToCLib')
 import Data.SBV.Compilers.CodeGen
@@ -84,7 +87,7 @@ sendRequestToSolver = Query.ask
 {- $coordinateSolverInfo
 In rare cases it might be necessary to send an arbitrary string down to the solver. Needless to say, this
 should be avoided if at all possible. Users should prefer the provided API. If you do find yourself
-needing 'send' and 'ask' directly, please get in touch to see if SBV can support a typed API for your use case.
+needing 'Data.SBV.Control.Utils.send' and 'Data.SBV.Control.Utils.ask' directly, please get in touch to see if SBV can support a typed API for your use case.
 Similarly, the function 'retrieveResponseFromSolver' might occasionally be necessary to clean-up the communication
 buffer. We would like to hear if you do need these functions regularly so we can provide better support.
 -}
