@@ -41,7 +41,7 @@ import Control.Concurrent (newEmptyMVar, takeMVar, putMVar, forkIO)
 import Control.DeepSeq    (NFData(..))
 import Control.Monad      (zipWithM)
 import Data.Char          (isSpace)
-import Data.Maybe         (fromMaybe)
+import Data.Maybe         (fromMaybe, isJust)
 import Data.Int           (Int8, Int16, Int32, Int64)
 import Data.List          (intercalate, isPrefixOf)
 import Data.Word          (Word8, Word16, Word32, Word64)
@@ -56,13 +56,13 @@ import System.Exit        (ExitCode(..))
 import System.IO          (hClose, hFlush, hPutStrLn, hGetContents, hGetLine)
 import System.Process     (runInteractiveProcess, waitForProcess, terminateProcess)
 
-import qualified Data.Map as M
+import qualified Data.Map.Strict as M
 
 import Data.SBV.Core.AlgReals
 import Data.SBV.Core.Data
 import Data.SBV.Core.Symbolic (SMTEngine, State(..))
 
-import Data.SBV.SMT.Utils     (showTimeoutValue, alignPlain, alignDiagnostic, debug, mergeSExpr, SMTException(..))
+import Data.SBV.SMT.Utils     (showTimeoutValue, alignPlain, debug, mergeSExpr, SBVException(..))
 
 import Data.SBV.Utils.PrettyNum
 import Data.SBV.Utils.Lib       (joinArgs, splitArgs)
@@ -73,30 +73,30 @@ import qualified System.Timeout as Timeout (timeout)
 
 -- | Extract the final configuration from a result
 resultConfig :: SMTResult -> SMTConfig
-resultConfig (Unsatisfiable c)   = c
+resultConfig (Unsatisfiable c _) = c
 resultConfig (Satisfiable   c _) = c
 resultConfig (SatExtField   c _) = c
 resultConfig (Unknown       c _) = c
 resultConfig (ProofError    c _) = c
 
--- | A 'prove' call results in a 'ThmResult'
+-- | A 'Data.SBV.prove' call results in a 'ThmResult'
 newtype ThmResult = ThmResult SMTResult
                   deriving NFData
 
--- | A 'sat' call results in a 'SatResult'
+-- | A 'Data.SBV.sat' call results in a 'SatResult'
 -- The reason for having a separate 'SatResult' is to have a more meaningful 'Show' instance.
 newtype SatResult = SatResult SMTResult
                   deriving NFData
 
--- | An 'allSat' call results in a 'AllSatResult'. The first boolean says whether we
+-- | An 'Data.SBV.allSat' call results in a 'AllSatResult'. The first boolean says whether we
 -- hit the max-model limit as we searched. The second boolean says whether
 -- there were prefix-existentials.
 newtype AllSatResult = AllSatResult (Bool, Bool, [SMTResult])
 
--- | A 'safe' call results in a 'SafeResult'
+-- | A 'Data.SBV.safe' call results in a 'SafeResult'
 newtype SafeResult   = SafeResult   (Maybe String, String, SMTResult)
 
--- | An 'optimize' call results in a 'OptimizeResult'. In the 'ParetoResult' case, the boolean is 'True'
+-- | An 'Data.SBV.optimize' call results in a 'OptimizeResult'. In the 'ParetoResult' case, the boolean is 'True'
 -- if we reached pareto-query limit and so there might be more unqueried results remaining. If 'False',
 -- it means that we have all the pareto fronts returned. See the 'Pareto' 'OptimizeStyle' for details.
 data OptimizeResult = LexicographicResult SMTResult
@@ -156,8 +156,8 @@ instance Show OptimizeResult where
                ParetoResult (False, [r]) -> sh (\s -> "Unique pareto front: " ++ s) r
                ParetoResult (False, rs)  -> multi "pareto optimal values" (zipWith shP [(1::Int)..] rs)
                ParetoResult (True,  rs)  ->    multi "pareto optimal values" (zipWith shP [(1::Int)..] rs)
-                                           ++ "\n*** Note: Pareto-front extraction was terminated before stream was ended as requested by the user."
-                                           ++ "\n***       There might be other (potentially infinitely more) results."
+                                           ++ "\n*** Note: Pareto-front extraction was terminated as requested by the user."
+                                           ++ "\n***       There might be many other results!"
 
        where multi w [] = "There are no " ++ w ++ " to display models for."
              multi _ xs = intercalate "\n" xs
@@ -172,7 +172,7 @@ instance Show OptimizeResult where
                                     (tag "Optimal in an extension field:" ++ "\n")
 
 -- | Instances of 'SatModel' can be automatically extracted from models returned by the
--- solvers. The idea is that the sbv infrastructure provides a stream of 'CW''s (constant-words)
+-- solvers. The idea is that the sbv infrastructure provides a stream of CW's (constant-words)
 -- coming from the solver, and the type @a@ is interpreted based on these constants. Many typical
 -- instances are already provided, so new instances can be declared with relative ease.
 --
@@ -256,7 +256,7 @@ instance SatModel Double where
   parseCWs (CW KDouble (CWDouble i) : r) = Just (i, r)
   parseCWs _                             = Nothing
 
--- | 'CW' as extracted from a model; trivial definition
+-- | @CW@ as extracted from a model; trivial definition
 instance SatModel CW where
   parseCWs (cw : r) = Just (cw, r)
   parseCWs []       = Nothing
@@ -349,7 +349,7 @@ class Modelable a where
   getModelObjectiveValue :: String -> a -> Maybe GeneralizedCW
   getModelObjectiveValue v r = v `M.lookup` getModelObjectives r
 
--- | Return all the models from an 'allSat' call, similar to 'extractModel' but
+-- | Return all the models from an 'Data.SBV.allSat' call, similar to 'extractModel' but
 -- is suitable for the case of multiple results.
 extractModels :: SatModel a => AllSatResult -> [a]
 extractModels (AllSatResult (_, _, xs)) = [ms | Right (_, ms) <- map getModelAssignment xs]
@@ -382,11 +382,11 @@ instance Modelable SatResult where
 
 -- | 'SMTResult' as a generic model provider
 instance Modelable SMTResult where
-  getModelAssignment (Unsatisfiable _) = Left "SBV.getModelAssignment: Unsatisfiable result"
-  getModelAssignment (Satisfiable _ m) = Right (False, parseModelOut m)
-  getModelAssignment (SatExtField _ _) = Left "SBV.getModelAssignment: The model is in an extension field"
-  getModelAssignment (Unknown _ m)     = Left $ "SBV.getModelAssignment: Solver state is unknown: " ++ m
-  getModelAssignment (ProofError _ s)  = error $ unlines $ "Backend solver complains: " : s
+  getModelAssignment (Unsatisfiable _ _) = Left "SBV.getModelAssignment: Unsatisfiable result"
+  getModelAssignment (Satisfiable _ m)   = Right (False, parseModelOut m)
+  getModelAssignment (SatExtField _ _)   = Left "SBV.getModelAssignment: The model is in an extension field"
+  getModelAssignment (Unknown _ m)       = Left $ "SBV.getModelAssignment: Solver state is unknown: " ++ show m
+  getModelAssignment (ProofError _ s)    = error $ unlines $ "Backend solver complains: " : s
 
   modelExists Satisfiable{}   = True
   modelExists Unknown{}       = False -- don't risk it
@@ -411,9 +411,9 @@ parseModelOut m = case parseCWs [c | (_, c) <- modelAssocs m] of
                    Just (_, ys) -> error $ "SBV.parseModelOut: Partially constructed model; remaining elements: " ++ show ys
                    Nothing      -> error $ "SBV.parseModelOut: Cannot construct a model from: " ++ show m
 
--- | Given an 'allSat' call, we typically want to iterate over it and print the results in sequence. The
--- 'displayModels' function automates this task by calling 'disp' on each result, consecutively. The first
--- 'Int' argument to 'disp' 'is the current model number. The second argument is a tuple, where the first
+-- | Given an 'Data.SBV.allSat' call, we typically want to iterate over it and print the results in sequence. The
+-- 'displayModels' function automates this task by calling @disp@ on each result, consecutively. The first
+-- 'Int' argument to @disp@ 'is the current model number. The second argument is a tuple, where the first
 -- element indicates whether the model is alleged (i.e., if the solver is not sure, returing Unknown)
 displayModels :: SatModel a => (Int -> (Bool, a) -> IO ()) -> AllSatResult -> IO Int
 displayModels disp (AllSatResult (_, _, ms)) = do
@@ -424,14 +424,16 @@ displayModels disp (AllSatResult (_, _, ms)) = do
 -- | Show an SMTResult; generic version
 showSMTResult :: String -> String -> String -> String -> String -> SMTResult -> String
 showSMTResult unsatMsg unkMsg satMsg satMsgModel satExtMsg result = case result of
-  Unsatisfiable _               -> unsatMsg
+  Unsatisfiable _ uc            -> unsatMsg ++ showUnsatCore uc
   Satisfiable _ (SMTModel _ []) -> satMsg
   Satisfiable _ m               -> satMsgModel ++ showModel cfg m
   SatExtField _ (SMTModel b _)  -> satExtMsg   ++ showModelDictionary cfg b
-  Unknown     _ r               -> unkMsg ++ ".\n" ++ "  Reason: " `alignPlain` r
+  Unknown     _ r               -> unkMsg ++ ".\n" ++ "  Reason: " `alignPlain` show r
   ProofError  _ []              -> "*** An error occurred. No additional information available. Try running in verbose mode"
   ProofError  _ ls              -> "*** An error occurred.\n" ++ intercalate "\n" (map ("***  " ++) ls)
  where cfg = resultConfig result
+       showUnsatCore Nothing   = ""
+       showUnsatCore (Just xs) = ". Unsat core:\n" ++ intercalate "\n" ["    " ++ x | x <- xs]
 
 -- | Show a model in human readable form. Ignore bindings to those variables that start
 -- with "__internal_sbv_" and also those marked as "nonModelVar" in the config; as these are only for internal purposes
@@ -489,12 +491,12 @@ pipeProcess cfg ctx execName opts pgm continuation = do
 
       Just execPath -> runSolver cfg ctx execPath opts pgm continuation
                        `C.catches`
-                        [ C.Handler (\(e :: SMTException)    -> C.throwIO e)
+                        [ C.Handler (\(e :: SBVException)    -> C.throwIO e)
                         , C.Handler (\(e :: C.ErrorCall)     -> C.throwIO e)
-                        , C.Handler (\(e :: C.SomeException) -> error $ unlines [ "Failed to start the external solver:\n" ++ show e
-                                                                                , "Make sure you can start " ++ show execPath
-                                                                                , "from the command line without issues."
-                                                                                ])
+                        , C.Handler (\(e :: C.SomeException) -> handleAsync e $ error $ unlines [ "Failed to start the external solver:\n" ++ show e
+                                                                                                , "Make sure you can start " ++ show execPath
+                                                                                                , "from the command line without issues."
+                                                                                                ])
                         ]
 
 -- | A standard engine interface. Most solvers follow-suit here in how we "chat" to them..
@@ -503,8 +505,8 @@ standardEngine :: String
                -> SMTEngine
 standardEngine envName envOptName cfg ctx pgm continuation = do
 
-    execName <-                    getEnv envName     `C.catch` (\(_ :: C.SomeException) -> return (executable (solver cfg)))
-    execOpts <- (splitArgs `fmap`  getEnv envOptName) `C.catch` (\(_ :: C.SomeException) -> return (options (solver cfg) cfg))
+    execName <-                    getEnv envName     `C.catch` (\(e :: C.SomeException) -> handleAsync e (return (executable (solver cfg))))
+    execOpts <- (splitArgs `fmap`  getEnv envOptName) `C.catch` (\(e :: C.SomeException) -> handleAsync e (return (options (solver cfg) cfg)))
 
     let cfg' = cfg {solver = (solver cfg) {executable = execName, options = const execOpts}}
 
@@ -530,7 +532,7 @@ data SolverLine = SolverRegular   String  -- ^ All is well
                 | SolverTimeout   String  -- ^ Timeout expired
                 | SolverException String  -- ^ Something else went wrong
 
--- | A variant of 'readProcessWithExitCode'; except it deals with SBV continuations
+-- | A variant of @readProcessWithExitCode@; except it deals with SBV continuations
 runSolver :: SMTConfig -> State -> FilePath -> [String] -> String -> (State -> IO a) -> IO a
 runSolver cfg ctx execPath opts pgm continuation
  = do let nm  = show (name (solver cfg))
@@ -610,7 +612,7 @@ runSolver cfg ctx execPath opts pgm continuation
 
 
                             go isFirst i sofar = do
-                                            errln <- safeGetLine isFirst outh `C.catch` (\(e :: C.SomeException) -> return (SolverException (show e)))
+                                            errln <- safeGetLine isFirst outh `C.catch` (\(e :: C.SomeException) -> handleAsync e (return (SolverException (show e))))
                                             case errln of
                                               SolverRegular ln -> let need  = i + parenDeficit ln
                                                                       -- make sure we get *something*
@@ -624,41 +626,45 @@ runSolver cfg ctx execPath opts pgm continuation
                                                                         (False, False) -> go False   need (ln:sofar)
                                                                         (False, True)  -> return (ln:sofar)
 
-                                              SolverException e -> error $ unlines $ [""
-                                                                                     , "*** Error     : " ++ e
-                                                                                     , "*** Executable: " ++ execPath
-                                                                                     , "*** Options   : " ++ joinArgs opts
-                                                                                     ]
-                                                                                  ++ [ "*** Request   : " `alignDiagnostic` command                 | Just command <- [mbCommand]]
-                                                                                  ++ [ "*** Response  : " `alignDiagnostic` unlines (reverse sofar) | not $ null sofar           ]
-                                                                                  ++ [ "***"
-                                                                                     , "*** Giving up!"
-                                                                                     ]
+                                              SolverException e -> do terminateProcess pid
+                                                                      C.throwIO SBVException { sbvExceptionDescription = e
+                                                                                             , sbvExceptionSent        = mbCommand
+                                                                                             , sbvExceptionExpected    = Nothing
+                                                                                             , sbvExceptionReceived    = Just $ unlines (reverse sofar)
+                                                                                             , sbvExceptionStdOut      = Nothing
+                                                                                             , sbvExceptionStdErr      = Nothing
+                                                                                             , sbvExceptionExitCode    = Nothing
+                                                                                             , sbvExceptionConfig      = cfg { solver = (solver cfg) { executable = execPath } }
+                                                                                             , sbvExceptionReason      = Nothing
+                                                                                             , sbvExceptionHint        = Nothing
+                                                                                             }
 
                                               SolverTimeout e -> do terminateProcess pid -- NB. Do not *wait* for the process, just quit.
-                                                                    error $ unlines $ [ ""
-                                                                                      , "*** Data.SBV: Timeout."
-                                                                                      , "***"
-                                                                                      , "***   " ++ e
-                                                                                      ]
-                                                                                   ++ [ "***   Response so far: " `alignDiagnostic` unlines (reverse sofar) | not $ null sofar]
-                                                                                   ++ [ "***   Last command sent was: " `alignDiagnostic` command | Just command <- [mbCommand]]
-                                                                                   ++ [ "***   Run with 'verbose=True' for further information" | not (verbose cfg)]
-                                                                                   ++ [ "***"
-                                                                                      , "*** Giving up!"
-                                                                                      ]
+                                                                    C.throwIO SBVException { sbvExceptionDescription = "Timeout! " ++ e
+                                                                                           , sbvExceptionSent        = mbCommand
+                                                                                           , sbvExceptionExpected    = Nothing
+                                                                                           , sbvExceptionReceived    = Just $ unlines (reverse sofar)
+                                                                                           , sbvExceptionStdOut      = Nothing
+                                                                                           , sbvExceptionStdErr      = Nothing
+                                                                                           , sbvExceptionExitCode    = Nothing
+                                                                                           , sbvExceptionConfig      = cfg { solver = (solver cfg) { executable = execPath } }
+                                                                                           , sbvExceptionReason      = Nothing
+                                                                                           , sbvExceptionHint        = if not (verbose cfg)
+                                                                                                                       then Just ["Run with 'verbose=True' for further information"]
+                                                                                                                       else Nothing
+                                                                                           }
 
                     terminateSolver = do hClose inh
                                          outMVar <- newEmptyMVar
-                                         out <- hGetContents outh `C.catch`  (\(e :: C.SomeException) -> return (show e))
+                                         out <- hGetContents outh `C.catch`  (\(e :: C.SomeException) -> handleAsync e (return (show e)))
                                          _ <- forkIO $ C.evaluate (length out) >> putMVar outMVar ()
-                                         err <- hGetContents errh `C.catch`  (\(e :: C.SomeException) -> return (show e))
+                                         err <- hGetContents errh `C.catch`  (\(e :: C.SomeException) -> handleAsync e (return (show e)))
                                          _ <- forkIO $ C.evaluate (length err) >> putMVar outMVar ()
                                          takeMVar outMVar
                                          takeMVar outMVar
-                                         hClose outh `C.catch`  (\(_ :: C.SomeException) -> return ())
-                                         hClose errh `C.catch`  (\(_ :: C.SomeException) -> return ())
-                                         ex <- waitForProcess pid `C.catch` (\(_ :: C.SomeException) -> return (ExitFailure (-999)))
+                                         hClose outh `C.catch`  (\(e :: C.SomeException) -> handleAsync e (return ()))
+                                         hClose errh `C.catch`  (\(e :: C.SomeException) -> handleAsync e (return ()))
+                                         ex <- waitForProcess pid `C.catch` (\(e :: C.SomeException) -> handleAsync e (return (ExitFailure (-999))))
                                          return (out, err, ex)
 
                     cleanUp
@@ -674,21 +680,26 @@ runSolver cfg ctx execPath opts pgm continuation
                              ExitSuccess -> return ()
                              _           -> if ignoreExitCode cfg
                                                then msg ["Ignoring non-zero exit code of " ++ show ex ++ " per user request!"]
-                                               else error $ unlines $  [ ""
-                                                                       , "*** Failed to complete the call to " ++ nm ++ ":"
-                                                                       , "*** Executable   : " ++ execPath
-                                                                       , "*** Options      : " ++ joinArgs opts
-                                                                       , "*** Exit code    : " ++ show ex
-                                                                       ]
-                                                                    ++ [ "*** Std-out      : " ++ intercalate "\n                   " (lines out)]
-                                                                    ++ [ "*** Std-err      : " ++ intercalate "\n                   " (lines err)]
-                                                                    ++ ["Giving up.."]
+                                               else C.throwIO SBVException { sbvExceptionDescription = "Failed to complete the call to " ++ nm
+                                                                           , sbvExceptionSent        = Nothing
+                                                                           , sbvExceptionExpected    = Nothing
+                                                                           , sbvExceptionReceived    = Nothing
+                                                                           , sbvExceptionStdOut      = Just out
+                                                                           , sbvExceptionStdErr      = Just err
+                                                                           , sbvExceptionExitCode    = Just ex
+                                                                           , sbvExceptionConfig      = cfg { solver = (solver cfg) { executable = execPath } }
+                                                                           , sbvExceptionReason      = Nothing
+                                                                           , sbvExceptionHint        = if not (verbose cfg)
+                                                                                                       then Just ["Run with 'verbose=True' for further information"]
+                                                                                                       else Nothing
+                                                                           }
+
                 return (send, ask, getResponseFromSolver, terminateSolver, cleanUp, pid)
 
       let executeSolver = do let sendAndGetSuccess :: Maybe Int -> String -> IO ()
                                  sendAndGetSuccess mbTimeOut l
                                    -- The pathetic case when the solver doesn't support queries, so we pretend it responded "success"
-                                   -- Currently ABC is the only such solver. Filed a request for ABC at: https://bitbucket.org/alanmi/abc/issues/70/
+                                   -- Currently ABC is the only such solver. Filed a request for ABC at: http://bitbucket.org/alanmi/abc/issues/70/
                                    | not (supportsCustomQueries (capabilities (solver cfg)))
                                    = do send mbTimeOut l
                                         debug cfg ["[ISSUE] " `alignPlain` l]
@@ -709,7 +720,8 @@ runSolver cfg ctx execPath opts pgm continuation
                                                                                     ]
 
                                                             -- put a sync point here before we die so we consume everything
-                                                            mbExtras <- (Right <$> getResponseFromSolver Nothing (Just 5000000)) `C.catch` (\(e :: C.SomeException) -> return (Left (show e)))
+                                                            mbExtras <- (Right <$> getResponseFromSolver Nothing (Just 5000000))
+                                                                        `C.catch` (\(e :: C.SomeException) -> handleAsync e (return (Left (show e))))
 
                                                             -- Ignore any exceptions from last sync, pointless.
                                                             let extras = case mbExtras of
@@ -720,16 +732,16 @@ runSolver cfg ctx execPath opts pgm continuation
                                                             let out = intercalate "\n" . lines $ outOrig
                                                                 err = intercalate "\n" . lines $ errOrig
 
-                                                                exc = SMTException { smtExceptionDescription = "Data.SBV: Unexpected non-success response from " ++ nm
-                                                                                   , smtExceptionSent        = l
-                                                                                   , smtExceptionExpected    = "success"
-                                                                                   , smtExceptionReceived    = r ++ "\n" ++ extras
-                                                                                   , smtExceptionStdOut      = out
-                                                                                   , smtExceptionStdErr      = Just err
-                                                                                   , smtExceptionExitCode    = Just ex
-                                                                                   , smtExceptionConfig      = cfg { solver = (solver cfg) {executable = execPath } }
-                                                                                   , smtExceptionReason      = Just reason
-                                                                                   , smtExceptionHint        = Nothing
+                                                                exc = SBVException { sbvExceptionDescription = "Unexpected non-success response from " ++ nm
+                                                                                   , sbvExceptionSent        = Just l
+                                                                                   , sbvExceptionExpected    = Just "success"
+                                                                                   , sbvExceptionReceived    = Just $ r ++ "\n" ++ extras
+                                                                                   , sbvExceptionStdOut      = Just out
+                                                                                   , sbvExceptionStdErr      = Just err
+                                                                                   , sbvExceptionExitCode    = Just ex
+                                                                                   , sbvExceptionConfig      = cfg { solver = (solver cfg) {executable = execPath } }
+                                                                                   , sbvExceptionReason      = Just reason
+                                                                                   , sbvExceptionHint        = Nothing
                                                                                    }
 
                                                             C.throwIO exc
@@ -782,6 +794,7 @@ runSolver cfg ctx execPath opts pgm continuation
                                                  , queryTerminate           = cleanUp
                                                  , queryTimeOutValue        = Nothing
                                                  , queryAssertionStackDepth = 0
+                                                 , queryTblArrPreserveIndex = Nothing
                                                  }
                                  qsp = queryState ctx
 
@@ -804,12 +817,12 @@ runSolver cfg ctx execPath opts pgm continuation
                             recordEndTime      cfg ctx
                             return r
 
-      launchSolver `C.catch` (\(e :: C.SomeException) -> do terminateProcess pid
-                                                            ec <- waitForProcess pid
-                                                            recordException    (transcript cfg) (show e)
-                                                            finalizeTranscript (transcript cfg) (Just ec)
-                                                            recordEndTime      cfg ctx
-                                                            C.throwIO e)
+      launchSolver `C.catch` (\(e :: C.SomeException) -> handleAsync e $ do terminateProcess pid
+                                                                            ec <- waitForProcess pid
+                                                                            recordException    (transcript cfg) (show e)
+                                                                            finalizeTranscript (transcript cfg) (Just ec)
+                                                                            recordEndTime      cfg ctx
+                                                                            C.throwIO e)
 
 -- | Compute and report the end time
 recordEndTime :: SMTConfig -> State -> IO ()
@@ -884,3 +897,13 @@ recordException (Just f) m = do ts <- show <$> getZonedTime
                         ++ [ ";;;"
                            , ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;"
                            ]
+
+-- We should not be catching/processing asynchronous exceptions.
+-- See http://github.com/LeventErkok/sbv/issues/410
+handleAsync :: C.SomeException -> IO a -> IO a
+handleAsync e cont
+  | isAsynchronous = C.throwIO e
+  | True           = cont
+  where -- Stealing this definition from the asynchronous exceptions package to reduce dependencies
+        isAsynchronous :: Bool
+        isAsynchronous = isJust (C.fromException e :: Maybe C.AsyncException) || isJust (C.fromException e :: Maybe C.SomeAsyncException)
