@@ -8,19 +8,22 @@
 --
 -- Conversion of symbolic programs to SMTLib format, Using v2 of the standard
 -----------------------------------------------------------------------------
+
 {-# LANGUAGE PatternGuards #-}
 
 module Data.SBV.SMT.SMTLib2(cvt, cvtInc) where
 
 import Data.Bits  (bit)
-import Data.List  (intercalate, partition)
+import Data.List  (intercalate, partition, unzip3)
+import Data.Maybe (listToMaybe, fromMaybe)
 
 import qualified Data.Foldable as F (toList)
-import qualified Data.Map      as M
-import qualified Data.IntMap   as IM
-import qualified Data.Set      as Set
+import qualified Data.Map.Strict      as M
+import qualified Data.IntMap.Strict   as IM
+import qualified Data.Set             as Set
 
 import Data.SBV.Core.Data
+import Data.SBV.Core.Kind (smtType)
 import Data.SBV.SMT.Utils
 import Data.SBV.Control.Types
 
@@ -35,10 +38,14 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
   where hasInteger     = KUnbounded `Set.member` kindInfo
         hasReal        = KReal      `Set.member` kindInfo
         hasFloat       = KFloat     `Set.member` kindInfo
+        hasString      = KString    `Set.member` kindInfo
+        hasChar        = KChar      `Set.member` kindInfo
         hasDouble      = KDouble    `Set.member` kindInfo
-        hasBVs         = not $ null [() | KBounded{} <- Set.toList kindInfo]
+        hasBVs         = hasChar || not (null [() | KBounded{} <- Set.toList kindInfo])   -- Remember, characters map to Word8
         usorts         = [(s, dt) | KUserSort s dt <- Set.toList kindInfo]
         hasNonBVArrays = (not . null) [() | (_, (_, (k1, k2), _)) <- arrs, not (isBounded k1 && isBounded k2)]
+        hasArrayInits  = (not . null) [() | (_, (_, _, ArrayFree (Just _))) <- arrs]
+        hasList        = any isList kindInfo
         rm             = roundingMode cfg
         solverCaps     = capabilities (solver cfg)
 
@@ -49,7 +56,7 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
                          []  -> Nothing
                          [l] -> Just l
                          ls  -> error $ unlines [ ""
-                                                , "*** Only one setOption call to 'SetLogic' is allowed, found: " ++ show (length ls)
+                                                , "*** Only one setOption call to 'setLogic' is allowed, found: " ++ show (length ls)
                                                 , "***  " ++ unwords (map show ls)
                                                 ]
            = case l of
@@ -58,6 +65,15 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
 
            -- Otherwise, we try to determine the most suitable logic.
            -- NB. This isn't really fool proof!
+
+           -- we never set QF_S (ALL seems to work better in all cases)
+
+           | hasArrayInits
+           = ["(set-logic ALL)"]
+
+           | hasString || hasList
+           = ["(set-logic ALL)"]
+
            | hasDouble || hasFloat
            = if hasInteger || not (null foralls)
              then ["(set-logic ALL)"]
@@ -81,7 +97,8 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
                     | True                     = "UF"
 
         -- SBV always requires the production of models!
-        getModels   = ["(set-option :produce-models true)"]
+        getModels   = "(set-option :produce-models true)"
+                    : concat [flattenConfig | hasList, Just flattenConfig <- [supportsFlattenedSequences solverCaps]]
 
         -- process all other settings we're given
         userSettings = concatMap opts $ solverSetOptions cfg
@@ -103,7 +120,7 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
              ++ [ "; --- optimization tracker variables ---" | not (null trackerVars) ]
              ++ [ "(declare-fun " ++ show s ++ " " ++ swFunType [] s ++ ") ; tracks " ++ nm | (s, nm) <- trackerVars]
              ++ [ "; --- constant tables ---" ]
-             ++ concatMap constTable constTables
+             ++ concatMap (constTable False) constTables
              ++ [ "; --- skolemized tables ---" ]
              ++ map (skolemTable (unwords (map swType foralls))) skolemTables
              ++ [ "; --- arrays ---" ]
@@ -121,6 +138,10 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
                 ]
              ++ map mkAssign postQuantifierAssigns
 
+             ++ concat arrayDelayeds
+
+             ++ concat arraySetups
+
              ++ delayedAsserts delayedEqualities
 
              ++ finalAssert
@@ -145,8 +166,8 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
 
         (constTables, skolemTables) = ([(t, d) | (t, Left d) <- allTables], [(t, d) | (t, Right d) <- allTables])
         allTables = [(t, genTableData rm skolemMap (not (null foralls), forallArgs) (map fst consts) t) | t <- tbls]
-        (arrayConstants, allArrayDelayeds) = unzip $ map (declArray (not (null foralls)) (map fst consts) skolemMap) arrs
-        delayedEqualities = concatMap snd skolemTables ++ concat allArrayDelayeds
+        (arrayConstants, arrayDelayeds, arraySetups) = unzip3 $ map (declArray cfg False (not (null foralls)) consts skolemMap) arrs
+        delayedEqualities = concatMap snd skolemTables
 
         delayedAsserts []              = []
         delayedAsserts ds@(deH : deTs)
@@ -157,17 +178,27 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
 
         finalAssert
           | null foralls
-          = map (\a -> "(assert " ++ named a ++ ")") assertions
+          =    map (\a -> "(assert "      ++ uncurry addAnnotations a ++ ")") hardAsserts
+            ++ map (\a -> "(assert-soft " ++ uncurry addAnnotations a ++ ")") softAsserts
           | not (null namedAsserts)
-          = error $ intercalate "\n" [ "SBV: Named constraints and quantifiers cannot be mixed!"
+          = error $ intercalate "\n" [ "SBV: Constraints with attributes and quantifiers cannot be mixed!"
                                      , "   Quantified variables: " ++ unwords (map show foralls)
                                      , "   Named constraints   : " ++ intercalate ", " (map show namedAsserts)
                                      ]
+          | not (null softAsserts)
+          = error $ intercalate "\n" [ "SBV: Soft constraints and quantifiers cannot be mixed!"
+                                     , "   Quantified variables: " ++ unwords (map show foralls)
+                                     , "   Soft constraints    : " ++ intercalate ", " (map show softAsserts)
+                                     ]
           | True
           = [impAlign (letShift combined) ++ replicate noOfCloseParens ')']
-          where namedAsserts = [n | (Just n, _) <- assertions]
+          where namedAsserts = [findName attrs | (_, attrs, _) <- assertions, not (null attrs)]
+                 where findName attrs = fromMaybe "<anonymous>" (listToMaybe [nm | (":named", nm) <- attrs])
 
-                combined = case map snd assertions of
+                hardAsserts = [(attr, v) | (False, attr, v) <- assertions]
+                softAsserts = [(attr, v) | (True,  attr, v) <- assertions]
+
+                combined = case map snd hardAsserts of
                              [x] -> x
                              xs  -> "(and " ++ unwords xs ++ ")"
 
@@ -190,13 +221,14 @@ cvt kindInfo isSat comments (inputs, trackerVars) skolemInps consts tbls arrs ui
         -- That is, we always assert all path constraints and path conditions AND
         --     -- negation of the output in a prove
         --     -- output itself in a sat
+        assertions :: [(Bool, [(String, String)], String)]
         assertions
-           | null finals = [(Nothing, cvtSW skolemMap trueSW)]
+           | null finals = [(False, [], cvtSW skolemMap trueSW)]
            | True        = finals
 
-           where finals  = cstrs' ++ maybe [] (\r -> [(Nothing, r)]) mbO
+           where finals  = cstrs' ++ maybe [] (\r -> [(False, [], r)]) mbO
 
-                 cstrs' =  [(mbNm, c') | (mbNm, c) <- cstrs, Just c' <- [pos c]]
+                 cstrs' =  [(isSoft, attrs, c') | (isSoft, attrs, c) <- cstrs, Just c' <- [pos c]]
 
                  mbO | isSat = pos out
                      | True  = neg out
@@ -242,8 +274,8 @@ declSort (s, Right fs) = [ "(declare-datatypes () ((" ++ s ++ " " ++ unwords (ma
               body (c:cs) i = "(ite (= x " ++ c ++ ") " ++ show i ++ " " ++ body cs (i+1) ++ ")"
 
 -- | Convert in a query context
-cvtInc :: SMTLibIncConverter [String]
-cvtInc inps ks consts arrs tbls (SBVPgm asgnsSeq) cfg =
+cvtInc :: Bool -> SMTLibIncConverter [String]
+cvtInc afterAPush inps ks consts arrs tbls uis (SBVPgm asgnsSeq) cfg =
             -- sorts
                concatMap declSort [(s, dt) | KUserSort s dt <- Set.toList ks]
             -- constants
@@ -252,12 +284,16 @@ cvtInc inps ks consts arrs tbls (SBVPgm asgnsSeq) cfg =
             ++ map declInp inps
             -- arrays
             ++ concat arrayConstants
+            -- uninterpreteds
+            ++ concatMap declUI uis
             -- tables
-            ++ concatMap constTable allTables
+            ++ concatMap (constTable afterAPush) allTables
             -- expressions
             ++ map  (declDef cfg skolemMap tableMap) (F.toList asgnsSeq)
             -- delayed equalities
-            ++ concatMap asrt arrayDelayeds
+            ++ concat arrayDelayeds
+            -- array setups
+            ++ concat arraySetups
   where -- NB. The below setting of skolemMap to empty is OK, since we do
         -- not support queries in the context of skolemized variables
         skolemMap = M.empty
@@ -266,14 +302,12 @@ cvtInc inps ks consts arrs tbls (SBVPgm asgnsSeq) cfg =
 
         declInp (s, _) = "(declare-fun " ++ show s ++ " () " ++ swType s ++ ")"
 
-        (arrayConstants, arrayDelayeds) = unzip $ map (declArray False (map fst consts) skolemMap) arrs
+        (arrayConstants, arrayDelayeds, arraySetups) = unzip3 $ map (declArray cfg afterAPush False consts skolemMap) arrs
 
         allTables = [(t, either id id (genTableData rm skolemMap (False, []) (map fst consts) t)) | t <- tbls]
         tableMap  = IM.fromList $ map mkTable allTables
           where mkTable (((t, _, _), _), _) = (t, "table" ++ show t)
 
-        asrt = map (\s -> "(assert " ++ s ++ ")")
-
 declDef :: SMTConfig -> SkolemMap -> TableMap -> (SW, SBVExpr) -> String
 declDef cfg skolemMap tableMap (s, expr) =
         case expr of
@@ -297,11 +331,29 @@ declUI (i, t) = ["(declare-fun " ++ i ++ " " ++ cvtType t ++ ")"]
 declAx :: (String, [String]) -> String
 declAx (nm, ls) = (";; -- user given axiom: " ++ nm ++ "\n") ++ intercalate "\n" ls
 
-constTable :: (((Int, Kind, Kind), [SW]), [String]) -> [String]
-constTable (((i, ak, rk), _elts), is) = decl : map wrap is
+constTable :: Bool -> (((Int, Kind, Kind), [SW]), [String]) -> [String]
+constTable afterAPush (((i, ak, rk), _elts), is) = decl : zipWith wrap [(0::Int)..] is ++ setup
   where t       = "table" ++ show i
         decl    = "(declare-fun " ++ t ++ " (" ++ smtType ak ++ ") " ++ smtType rk ++ ")"
-        wrap  s = "(assert " ++ s ++ ")"
+
+        -- Arrange for initializers
+        mkInit idx   = "table" ++ show i ++ "_initializer_" ++ show (idx :: Int)
+        initializer  = "table" ++ show i ++ "_initializer"
+        wrap index s
+          | afterAPush = "(define-fun " ++ mkInit index ++ " () Bool " ++ s ++ ")"
+          | True       = "(assert " ++ s ++ ")"
+        lis          = length is
+        setup
+          | not afterAPush = []
+          | lis == 0       = [ "(define-fun " ++ initializer ++ " () Bool true) ; no initializiation needed"
+                             ]
+          | lis == 1       = [ "(define-fun " ++ initializer ++ " () Bool " ++ mkInit 0 ++ ")"
+                             , "(assert " ++ initializer ++ ")"
+                             ]
+          | True           = [ "(define-fun " ++ initializer ++ " () Bool (and " ++ unwords (map mkInit [0..lis - 1]) ++ "))"
+                             , "(assert " ++ initializer ++ ")"
+                             ]
+
 
 skolemTable :: String -> (((Int, Kind, Kind), [SW]), [String]) -> String
 skolemTable qsIn (((i, ak, rk), _elts), _) = decl
@@ -328,25 +380,55 @@ genTableData rm skolemMap (_quantified, args) consts ((i, aknd, _), elts)
 -- we might have to skolemize those. Implement this properly.
 -- The difficulty is with the Mutate/Merge: We have to postpone an init if
 -- the components are themselves postponed, so this cannot be implemented as a simple map.
-declArray :: Bool -> [SW] -> SkolemMap -> (Int, ArrayInfo) -> ([String], [String])
-declArray quantified consts skolemMap (i, (_, (aKnd, bKnd), ctx)) = (adecl : map (wrap . snd) pre, map snd post)
-  where topLevel = not quantified || case ctx of
-                                       ArrayFree         -> True
-                                       ArrayMutate _ a b -> all (`elem` consts) [a, b]
-                                       ArrayMerge c _ _  -> c `elem` consts
+declArray :: SMTConfig -> Bool -> Bool -> [(SW, CW)] -> SkolemMap -> (Int, ArrayInfo) -> ([String], [String], [String])
+declArray cfg afterAPush quantified consts skolemMap (i, (_, (aKnd, bKnd), ctx)) = (adecl : zipWith wrap [(0::Int)..] (map snd pre), zipWith wrap [lpre..] (map snd post), setup)
+  where constNames = map fst consts
+        topLevel = not quantified || case ctx of
+                                       ArrayFree mbi      -> maybe True (`elem` constNames) mbi
+                                       ArrayMutate _ a b  -> all (`elem` constNames) [a, b]
+                                       ArrayMerge c _ _   -> c `elem` constNames
         (pre, post) = partition fst ctxInfo
         nm = "array_" ++ show i
         ssw sw
-         | topLevel || sw `elem` consts
+         | topLevel || sw `elem` constNames
          = cvtSW skolemMap sw
          | True
          = tbd "Non-constant array initializer in a quantified context"
-        adecl = "(declare-fun " ++ nm ++ " () (Array " ++ smtType aKnd ++ " " ++ smtType bKnd ++ "))"
+        atyp  = "(Array " ++ smtType aKnd ++ " " ++ smtType bKnd ++ ")"
+
+        adecl = case ctx of
+                  ArrayFree (Just v) -> "(define-fun "  ++ nm ++ " () " ++ atyp ++ " ((as const " ++ atyp ++ ") " ++ constInit v ++ "))"
+                  _                  -> "(declare-fun " ++ nm ++ " () " ++ atyp ++                                                  ")"
+
+        -- CVC4 chokes if the initializer is not a constant. (Z3 is ok with it.) So, print it as
+        -- a constant if we have it in the constants; otherwise, we merely print it and hope for the best.
+        constInit v = case v `lookup` consts of
+                        Nothing -> ssw v                      -- Z3 will work, CVC4 will choke. Others don't even support this.
+                        Just c  -> cvtCW (roundingMode cfg) c -- Z3 and CVC4 will work. Other's don't support this.
+
         ctxInfo = case ctx of
-                    ArrayFree         -> []
-                    ArrayMutate j a b -> [(all (`elem` consts) [a, b], "(= " ++ nm ++ " (store array_" ++ show j ++ " " ++ ssw a ++ " " ++ ssw b ++ "))")]
-                    ArrayMerge  t j k -> [(t `elem` consts,            "(= " ++ nm ++ " (ite " ++ ssw t ++ " array_" ++ show j ++ " array_" ++ show k ++ "))")]
-        wrap s = "(assert " ++ s ++ ")"
+                    ArrayFree _       -> []
+                    ArrayMutate j a b -> [(all (`elem` constNames) [a, b], "(= " ++ nm ++ " (store array_" ++ show j ++ " " ++ ssw a ++ " " ++ ssw b ++ "))")]
+                    ArrayMerge  t j k -> [(t `elem` constNames,            "(= " ++ nm ++ " (ite " ++ ssw t ++ " array_" ++ show j ++ " array_" ++ show k ++ "))")]
+
+        -- Arrange for initializers
+        mkInit idx    = "array_" ++ show i ++ "_initializer_" ++ show (idx :: Int)
+        initializer   = "array_" ++ show i ++ "_initializer"
+        wrap index s
+          | afterAPush = "(define-fun " ++ mkInit index ++ " () Bool " ++ s ++ ")"
+          | True       = "(assert " ++ s ++ ")"
+        lpre          = length pre
+        lAll          = lpre + length post
+        setup
+          | not afterAPush = []
+          | lAll == 0      = [ "(define-fun " ++ initializer ++ " () Bool true) ; no initializiation needed"
+                             ]
+          | lAll == 1      = [ "(define-fun " ++ initializer ++ " () Bool " ++ mkInit 0 ++ ")"
+                             , "(assert " ++ initializer ++ ")"
+                             ]
+          | True           = [ "(define-fun " ++ initializer ++ " () Bool (and " ++ unwords (map mkInit [0..lAll - 1]) ++ "))"
+                             , "(assert " ++ initializer ++ ")"
+                             ]
 
 swType :: SW -> String
 swType s = smtType (kindOf s)
@@ -354,15 +436,6 @@ swType s = smtType (kindOf s)
 swFunType :: [SW] -> SW -> String
 swFunType ss s = "(" ++ unwords (map swType ss) ++ ") " ++ swType s
 
-smtType :: Kind -> String
-smtType KBool           = "Bool"
-smtType (KBounded _ sz) = "(_ BitVec " ++ show sz ++ ")"
-smtType KUnbounded      = "Int"
-smtType KReal           = "Real"
-smtType KFloat          = "(_ FloatingPoint  8 24)"
-smtType KDouble         = "(_ FloatingPoint 11 53)"
-smtType (KUserSort s _) = s
-
 cvtType :: SBVType -> String
 cvtType (SBVType []) = error "SBV.SMT.SMTLib2.cvtType: internal: received an empty type!"
 cvtType (SBVType xs) = "(" ++ unwords (map smtType body) ++ ") " ++ smtType ret
@@ -398,6 +471,9 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
         doubleOp = any isDouble  arguments
         floatOp  = any isFloat   arguments
         boolOp   = all isBoolean arguments
+        charOp   = any isChar    arguments
+        stringOp = any isString  arguments
+        listOp   = any isList    arguments
 
         bad | intOp = error $ "SBV.SMTLib2: Unsupported operation on unbounded integers: " ++ show expr
             | True  = error $ "SBV.SMTLib2: Unsupported operation on real values: " ++ show expr
@@ -470,6 +546,22 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
           = let idx v = "(" ++ s ++ "_constrIndex " ++ v ++ ")" in "(" ++ o ++ " " ++ idx a ++ " " ++ idx b ++ ")"
         unintComp o sbvs = error $ "SBV.SMT.SMTLib2.sh.unintComp: Unexpected arguments: "   ++ show (o, sbvs)
 
+        -- NB. String comparisons are currently not supported by Z3; but will be with the new logic.
+        stringCmp swap o [a, b]
+          | KString <- kindOf (head arguments)
+          = let [a1, a2] | swap = [b, a]
+                         | True = [a, b]
+            in "(" ++ o ++ " " ++ a1 ++ " " ++ a2 ++ ")"
+        stringCmp _ o sbvs = error $ "SBV.SMT.SMTLib2.sh.stringCmp: Unexpected arguments: " ++ show (o, sbvs)
+
+        -- NB. Likewise for sequences
+        seqCmp swap o [a, b]
+          | KList{} <- kindOf (head arguments)
+          = let [a1, a2] | swap = [b, a]
+                         | True = [a, b]
+            in "(" ++ o ++ " " ++ a1 ++ " " ++ a2 ++ ")"
+        seqCmp _ o sbvs = error $ "SBV.SMT.SMTLib2.sh.seqCmp: Unexpected arguments: " ++ show (o, sbvs)
+
         lift1  o _ [x]    = "(" ++ o ++ " " ++ x ++ ")"
         lift1  o _ sbvs   = error $ "SBV.SMT.SMTLib2.sh.lift1: Unexpected arguments: "   ++ show (o, sbvs)
 
@@ -485,6 +577,9 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                               KReal         -> error "SBV.SMT.SMTLib2.cvtExp: unexpected real valued index"
                               KFloat        -> error "SBV.SMT.SMTLib2.cvtExp: unexpected float valued index"
                               KDouble       -> error "SBV.SMT.SMTLib2.cvtExp: unexpected double valued index"
+                              KChar         -> error "SBV.SMT.SMTLib2.cvtExp: unexpected char valued index"
+                              KString       -> error "SBV.SMT.SMTLib2.cvtExp: unexpected string valued index"
+                              KList k       -> error $ "SBV.SMT.SMTLib2.cvtExp: unexpected list valued: " ++ show k
                               KUserSort s _ -> error $ "SBV.SMT.SMTLib2.cvtExp: unexpected uninterpreted valued index: " ++ s
                 lkUp = "(" ++ getTable tableMap t ++ " " ++ ssw i ++ ")"
                 cond
@@ -497,6 +592,9 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                                 KReal         -> ("<", "<=")
                                 KFloat        -> ("fp.lt", "fp.leq")
                                 KDouble       -> ("fp.lt", "fp.geq")
+                                KChar         -> error "SBV.SMT.SMTLib2.cvtExp: unexpected string valued index"
+                                KString       -> error "SBV.SMT.SMTLib2.cvtExp: unexpected string valued index"
+                                KList k       -> error $ "SBV.SMT.SMTLib2.cvtExp: unexpected sequence valued index: " ++ show k
                                 KUserSort s _ -> error $ "SBV.SMT.SMTLib2.cvtExp: unexpected uninterpreted valued index: " ++ s
                 mkCnst = cvtCW rm . mkConstCW (kindOf i)
                 le0  = "(" ++ less ++ " " ++ ssw i ++ " " ++ mkCnst 0 ++ ")"
@@ -541,7 +639,7 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                                , (Join, lift2 "concat")
                                ]
 
-        sh (SBVApp (Label m) [a]) = curry named (Just m) $ cvtSW skolemMap a  -- This won't be reached; but just in case!
+        sh (SBVApp (Label _) [a]) = cvtSW skolemMap a  -- This won't be reached; but just in case!
 
         sh (SBVApp (IEEEFP (FP_Cast kFrom kTo m)) args) = handleFPCast kFrom kTo (ssw m) (unwords (map ssw args))
         sh (SBVApp (IEEEFP w                    ) args) = "(" ++ show w ++ " " ++ unwords (map ssw args) ++ ")"
@@ -551,6 +649,15 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
           | True       = reducePB pb args'
           where args' = map ssw args
 
+        -- NB: Z3 semantics have the predicates reversed: i.e., it returns true if overflow isn't possible. Hence the not.
+        sh (SBVApp (OverflowOp op) args) = "(not (" ++ show op ++ " " ++ unwords (map ssw args) ++ "))"
+
+        -- Note the unfortunate reversal in StrInRe..
+        sh (SBVApp (StrOp (StrInRe r)) args) = "(str.in.re " ++ unwords (map ssw args) ++ " " ++ show r ++ ")"
+        sh (SBVApp (StrOp op)          args) = "(" ++ show op ++ " " ++ unwords (map ssw args) ++ ")"
+
+        sh (SBVApp (SeqOp op) args) = "(" ++ show op ++ " " ++ unwords (map ssw args) ++ ")"
+
         sh inp@(SBVApp op args)
           | intOp, Just f <- lookup op smtOpIntTable
           = f True (map ssw args)
@@ -562,6 +669,12 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
           = f (any hasSign args) (map ssw args)
           | floatOp || doubleOp, Just f <- lookup op smtOpFloatDoubleTable
           = f (any hasSign args) (map ssw args)
+          | charOp, Just f <- lookup op smtCharTable
+          = f False (map ssw args)
+          | stringOp, Just f <- lookup op smtStringTable
+          = f (map ssw args)
+          | listOp, Just f <- lookup op smtListTable
+          = f (map ssw args)
           | Just f <- lookup op uninterpretedTable
           = f (map ssw args)
           | True
@@ -613,7 +726,7 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                                     , (LessEq,        lift2Cmp  "<=" "fp.leq")
                                     , (GreaterEq,     lift2Cmp  ">=" "fp.geq")
                                     ]
-                -- equality and comparisons are the only thing that works on uninterpreted sorts
+                -- equality and comparisons are the only thing that works on uninterpreted sorts and pretty much everything else
                 uninterpretedTable = [ (Equal,       lift2S "="        "="        True)
                                      , (NotEqual,    liftNS "distinct" "distinct" True)
                                      , (LessThan,    unintComp "<")
@@ -621,6 +734,33 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
                                      , (LessEq,      unintComp "<=")
                                      , (GreaterEq,   unintComp ">=")
                                      ]
+                -- For chars, the underlying type is currently SWord8, so we go with the regular bit-vector operations
+                -- TODO: This will change when we move to unicode!
+                smtCharTable = [ (Equal,         eqBV)
+                               , (NotEqual,      neqBV)
+                               , (LessThan,      lift2S  "bvult" (error "smtChar.<: did-not expect signed char here!"))
+                               , (GreaterThan,   lift2S  "bvugt" (error "smtChar.>: did-not expect signed char here!"))
+                               , (LessEq,        lift2S  "bvule" (error "smtChar.<=: did-not expect signed char here!"))
+                               , (GreaterEq,     lift2S  "bvuge" (error "smtChar.>=: did-not expect signed char here!"))
+                               ]
+                -- For strings, equality and comparisons are the only operators
+                -- TODO: The string comparison operators will most likely change with the new theory!
+                smtStringTable = [ (Equal,       lift2S "="        "="        True)
+                                 , (NotEqual,    liftNS "distinct" "distinct" True)
+                                 , (LessThan,    stringCmp False "str.<")
+                                 , (GreaterThan, stringCmp True  "str.<")
+                                 , (LessEq,      stringCmp False "str.<=")
+                                 , (GreaterEq,   stringCmp True  "str.<=")
+                                 ]
+                -- For lists, equality is really the only operator
+                -- Likewise here, things might change for comparisons
+                smtListTable = [ (Equal,       lift2S "="        "="        True)
+                               , (NotEqual,    liftNS "distinct" "distinct" True)
+                               , (LessThan,    seqCmp False "seq.<")
+                               , (GreaterThan, seqCmp True  "seq.<")
+                               , (LessEq,      seqCmp False "seq.<=")
+                               , (GreaterEq,   seqCmp True  "seq.<=")
+                               ]
 
 -----------------------------------------------------------------------------------------------
 -- Casts supported by SMTLib. (From: <http://smtlib.cs.uiowa.edu/theories-FloatingPoint.shtml>)
@@ -640,7 +780,7 @@ cvtExp caps rm skolemMap tableMap expr@(SBVApp _ arguments) = sh expr
 --   ((_ fp.to_ubv m) RoundingMode (_ FloatingPoint eb sb) (_ BitVec m))
 --
 --   ; to signed machine integer, represented as a 2's complement bit vector
---   ((_ fp.to_sbv m) RoundingMode (_ FloatingPoint eb sb) (_ BitVec m)) 
+--   ((_ fp.to_sbv m) RoundingMode (_ FloatingPoint eb sb) (_ BitVec m))
 --
 --   ; to real
 --   (fp.to_real (_ FloatingPoint eb sb) Real)
@@ -732,6 +872,14 @@ handleKindCast kFrom kTo a
          | m == n = a
          | True   = extract (n - 1)
 
+        -- NB. The following works regardless n < 0 or not, because the first thing we
+        -- do is to compute "reduced" to bring it down to the correct range. It also works
+        -- regardless were mapping to signed or unsigned bit-vector; because the representation
+        -- is the same.
+        --
+        -- NB2. (TODO) There is an SMTLib equivalent of this function, called int2bv. However, it
+        -- used to be uninterpreted for a long time by Z3; though I think that got fixed. We
+        -- might want to simply use that if it's reliably available across the board in solvers.
         i2b n = "(let (" ++ reduced ++ ") (let (" ++ defs ++ ") " ++ body ++ "))"
           where b i      = show (bit i :: Integer)
                 reduced  = "(__a (mod " ++ a ++ " " ++ b n ++ "))"
@@ -774,8 +922,3 @@ reducePB op args = case op of
 
   where addIf :: [Int] -> String
         addIf cs = "(+ " ++ unwords ["(ite " ++ a ++ " " ++ show c ++ " 0)" | (a, c) <- zip args cs] ++ ")"
-
--- Add a named annotation
-named :: (Maybe String, String) -> String
-named (Nothing, x) = x
-named (Just nm, x) = annotateWithName nm x
