@@ -25,8 +25,8 @@ fpRatio0 x
 
 -- | The SMT-Lib (in particular Z3) implementation for min/max for floats does not agree with
 -- Haskell's; and also it does not agree with what the hardware does. Sigh.. See:
---      <https://ghc.haskell.org/trac/ghc/ticket/10378>
---      <https://github.com/Z3Prover/z3/issues/68>
+--      <http://ghc.haskell.org/trac/ghc/ticket/10378>
+--      <http://github.com/Z3Prover/z3/issues/68>
 -- So, we codify here what the Z3 (SMTLib) is implementing for fpMax.
 -- The discrepancy with Haskell is that the NaN propagation doesn't work in Haskell
 -- The discrepancy with x86 is that given +0/-0, x86 returns the second argument; SMTLib is non-deterministic
@@ -40,7 +40,7 @@ fpMaxH x y
    where isN0   = isNegativeZero
          isP0 a = a == 0 && not (isN0 a)
 
--- | SMTLib compliant definition for 'fpMin'. See the comments for 'fpMax'.
+-- | SMTLib compliant definition for 'Data.SBV.fpMin'. See the comments for 'Data.SBV.fpMax'.
 fpMinH :: RealFloat a => a -> a -> a
 fpMinH x y
    | isNaN x                                  = y
@@ -102,6 +102,20 @@ fpIsEqualObjectH a b
   | isNegativeZero b = isNegativeZero a
   | True             = a == b
 
+-- | Ordering for floats, avoiding the +0/-0/NaN issues. Note that this is
+-- essentially used for indexing into a map, so we need to be total. Thus,
+-- the order we pick is:
+--    NaN -oo -0 +0 +oo
+-- The placement of NaN here is questionable, but immaterial.
+fpCompareObjectH :: RealFloat a => a -> a -> Ordering
+fpCompareObjectH a b
+  | a `fpIsEqualObjectH` b   = EQ
+  | isNaN a                  = LT
+  | isNaN b                  = GT
+  | isNegativeZero a, b == 0 = LT
+  | isNegativeZero b, a == 0 = GT
+  | True                     = a `compare` b
+
 -- | Check if a number is "normal." Note that +0/-0 is not considered a normal-number
 -- and also this is not simply the negation of isDenormalized!
 fpIsNormalizedH :: RealFloat a => a -> Bool
