@@ -6,12 +6,14 @@
 -- Maintainer  :  erkokl@gmail.com
 -- Stability   :  experimental
 --
--- Test suite for Data.SBV.Examples.Puzzles.U2Bridge
+-- Test suite for Documentation.SBV.Examples.Puzzles.U2Bridge
 -----------------------------------------------------------------------------
 
 module TestSuite.Puzzles.U2Bridge(tests) where
 
-import Data.SBV.Examples.Puzzles.U2Bridge
+import Documentation.SBV.Examples.Puzzles.U2Bridge
+
+import Data.List(sortOn)
 
 import Utils.SBVTestFramework
 
@@ -28,4 +30,5 @@ tests =
     ]
  where act     = do b <- exists_; p1 <- exists_; p2 <- exists_; return (b, p1, p2)
        count n = numberOfModels $ isValid `fmap` mapM (const act) [1..(n::Int)]
-       slv n   = sat $ isValid `fmap` mapM (const act) [1..(n::Int)]
+       slv n   = rearrange `fmap` allSat (isValid `fmap` mapM (const act) [1..(n::Int)])
+       rearrange (AllSatResult (b1, b2, ms)) = AllSatResult (b1, b2, sortOn (show . SatResult) ms)
